class BaseOptimizer:
    def __init__(self, space, params):
        raise NotImplementedError()

    def solve(model):
        raise NotImplementedError()

    def sample_feas(num_points):
        raise NotImplementedError()



from entmoot.models.uncertainty_models.base_distance import NonCatDistance
import numpy as np


class L1Distance(NonCatDistance):
    def _get_distance(self, x_left, x_right):
        if x_left.ndim == 1:
            dist = np.sum(np.abs(x_left - x_right))
        else:
            dist = np.sum(np.abs(x_left - x_right), axis=1)
        return dist

    def _array_predict(self, X):
        raise NotImplementedError()

    def get_gurobipy_model_constr_terms(self, model):
        from gurobipy import GRB, quicksum

        feat = model._all_feat

        # define auxiliary variables
        feat_dict = {
            idx: self._problem_config.feat_list[idx]
            for idx in self._problem_config.non_cat_idx
        }

        aux_pos = model.addVars(len(self.x_trafo), feat_dict, name="aux_pos", vtype="C")

        aux_neg = model.addVars(len(self.x_trafo), feat_dict, name="aux_neg", vtype="C")

        # define distance constraints
        constr_list = []
        for data_idx, xi in enumerate(self.x_trafo):
            for i, idx in enumerate(self._problem_config.non_cat_idx):
                # capture positive and negative contributions
                model.addConstr(
                    (xi[i] - (feat[idx] - self.shift[i]) / self.scale[i])
                    == aux_pos[data_idx, idx] - aux_neg[data_idx, idx],
                    name=f"unc_aux_({data_idx},{idx})",
                )

                # make sure only one variable is non-zero
                model.addSOS(
                    GRB.SOS_TYPE1, [aux_pos[data_idx, idx], aux_neg[data_idx, idx]]
                )

            # define individual sum terms
            constr = quicksum(
                aux_pos[data_idx, idx] + aux_neg[data_idx, idx]
                for idx in self._problem_config.non_cat_idx
            )

            constr_list.append(constr)

        model.update()
        return constr_list

    def get_pyomo_model_constr_terms(self, model):
        import pyomo.environ as pyo

        feat = model._all_feat

        # define auxiliary variables
        feat_dict = {
            idx: self._problem_config.feat_list[idx]
            for idx in self._problem_config.non_cat_idx
        }

        indices_aux_vars = [(i, j) for i in range(len(self.x_trafo)) for j in feat_dict]
        model.aux_pos = pyo.Var(indices_aux_vars, domain=pyo.NonNegativeReals)
        model.aux_neg = pyo.Var(indices_aux_vars, domain=pyo.NonNegativeReals)

        indices_l1_constraints = [
            (data_idx, i, idx)
            for data_idx, _ in enumerate(self.x_trafo)
            for i, idx in enumerate(self._problem_config.non_cat_idx)
        ]

        def rule_contrs_l1_pos_neg_contr(modelobj, data_idx, i, idx):
            xi = self.x_trafo[data_idx]
            return (
                xi[i] - (feat[idx] - self.shift[i]) / self.scale[i]
            ) == modelobj.aux_pos[data_idx, idx] - modelobj.aux_neg[data_idx, idx]

        model.contrs_l1_pos_neg_contr = pyo.Constraint(
            indices_l1_constraints, rule=rule_contrs_l1_pos_neg_contr
        )

        model.l1_bigm = {
            (data_idx, idx): max(
                abs(
                    self.x_trafo[data_idx][i]
                    - (feat[idx].ub - self.shift[i]) / self.scale[i]
                ),
                abs(
                    self.x_trafo[data_idx][i]
                    - (feat[idx].lb - self.shift[i]) / self.scale[i]
                ),
            )
            for (data_idx, i, idx) in indices_l1_constraints
        }

        model.indices_bigm = list(model.l1_bigm.keys())

        model.y_pos_bigm = pyo.Var(model.indices_bigm, domain=pyo.Binary)
        model.y_neg_bigm = pyo.Var(model.indices_bigm, domain=pyo.Binary)

        def bigm_ub_pos(modelobj, data_idx, idx):
            return (
                modelobj.aux_pos[data_idx, idx]
                <= model.l1_bigm[data_idx, idx] * model.y_pos_bigm[data_idx, idx]
            )

        def bigm_ub_neg(modelobj, data_idx, idx):
            return (
                modelobj.aux_neg[data_idx, idx]
                <= model.l1_bigm[data_idx, idx] * model.y_neg_bigm[data_idx, idx]
            )

        model.constrs_l1_bigm_ub_pos = pyo.Constraint(
            model.indices_bigm, rule=bigm_ub_pos
        )
        model.constrs_l1_bigm_ub_neg = pyo.Constraint(
            model.indices_bigm, rule=bigm_ub_neg
        )

        def binaries_mutually_exclusive(modelobj, data_idx, idx):
            return (
                model.y_pos_bigm[data_idx, idx] + model.y_neg_bigm[data_idx, idx] == 1
            )

        model.constrs_l1_bigm_mut_excl = pyo.Constraint(
            model.indices_bigm, rule=binaries_mutually_exclusive
        )

        constr_list = [
            sum(
                model.aux_pos[data_idx, idx] + model.aux_neg[data_idx, idx]
                for idx in self._problem_config.non_cat_idx
            )
            for (data_idx, _) in enumerate(self.x_trafo)
        ]

        return constr_list



from entmoot.models.uncertainty_models.base_distance import CatDistance


class OverlapDistance(CatDistance):
    def _sim_mat_rule(self, x_left, x_right, cat_idx):
        return 1.0 if x_left == x_right else 0.0



from entmoot.problem_config import ProblemConfig
from entmoot.models.enting import Enting
from entmoot.optimizers.gurobi_opt import GurobiOptimizer
from entmoot.optimizers.pyomo_opt import PyomoOptimizer
from entmoot.models.model_params import EntingParams, UncParams, TreeTrainParams, TrainParams


import math

from entmoot import Enting, ProblemConfig, GurobiOptimizer, PyomoOptimizer
from entmoot.models.model_params import EntingParams, UncParams
from entmoot.benchmarks import (
    build_multi_obj_categorical_problem,
    eval_multi_obj_cat_testfunc,
)

import numpy as np
import pytest
import random
import pyomo.environ # noqa: F401

@pytest.mark.pipeline_test
def test_core_model_copy():
    # define problem
    problem_config = ProblemConfig(rnd_seed=73)
    # number of objectives
    number_objectives = 2
    build_multi_obj_categorical_problem(problem_config, n_obj=number_objectives)

    core_model_gurobi = problem_config.get_gurobi_model_core()
    core_model_gurobi_copy = problem_config.copy_gurobi_model_core(core_model_gurobi)

    assert len(core_model_gurobi.getVars()) == len(core_model_gurobi_copy.getVars())
    assert len(core_model_gurobi._all_feat) == len(core_model_gurobi_copy._all_feat)

    core_model_pyomo = problem_config.get_pyomo_model_core()
    core_model_pyomo_copy = problem_config.copy_pyomo_model_core(core_model_pyomo)

    assert len(core_model_pyomo.x) == len(core_model_pyomo_copy.x)
    assert len(core_model_pyomo._all_feat) == len(core_model_pyomo_copy._all_feat)


# @pytest.mark.pipeline_test
def test_multiobj_constraints():
    # define problem
    problem_config = ProblemConfig(rnd_seed=73)
    # number of objectives
    number_objectives = 2
    build_multi_obj_categorical_problem(problem_config, n_obj=number_objectives)

    # sample data
    rnd_sample = problem_config.get_rnd_sample_list(num_samples=20)
    testfunc_evals = eval_multi_obj_cat_testfunc(rnd_sample, n_obj=number_objectives)

    params = EntingParams(unc_params=UncParams(
        dist_metric="l1",
        acq_sense="exploration"
    ))
    enting = Enting(problem_config, params=params)
    # fit tree ensemble
    enting.fit(rnd_sample, testfunc_evals)

    # Add constraints

    # Gurobi version
    # get optimization model
    model_gur = problem_config.get_gurobi_model_core()
    # extract decision variables
    x = model_gur._all_feat[3]
    y = model_gur._all_feat[4]
    z = model_gur._all_feat[5]
    # add constraint that all variables should coincide
    model_gur.addConstr(x == y)
    model_gur.addConstr(y == z)
    # Update model
    model_gur.update()

    # Build GurobiOptimizer object and solve optimization problem
    params_gurobi = {"MIPGap": 1e-3}
    opt_gur = GurobiOptimizer(problem_config, params=params_gurobi)

    res_gur = opt_gur.solve(enting, model_core=model_gur)
    x_opt, y_opt, z_opt = res_gur.opt_point[3:]

    assert round(x_opt, 5) == round(y_opt, 5) and round(y_opt, 5) == round(z_opt, 5)

    # Pyomo version
    import pyomo.environ as pyo

    model_pyo = problem_config.get_pyomo_model_core()
    # extract decision variables
    x = model_pyo._all_feat[3]
    y = model_pyo._all_feat[4]
    z = model_pyo._all_feat[5]
    # add constraint that all variables should coincide
    model_pyo.xy_equal_constr = pyo.Constraint(expr=x == y)
    model_pyo.yz_equal_constr = pyo.Constraint(expr=y == z)

    # Build GurobiOptimizer object and solve optimization problem
    params_pyomo = {"solver_name": "gurobi"}
    opt_pyo = PyomoOptimizer(problem_config, params=params_pyomo)

    res_pyo = opt_pyo.solve(enting, model_core=model_pyo)
    x_opt, y_opt, z_opt = res_pyo.opt_point[3:]

    assert round(x_opt, 5) == round(y_opt, 5) and round(y_opt, 5) == round(z_opt, 5)


# @pytest.mark.pipeline_test
def test_simple_test():
    def my_func(x: float) -> float:
        return x**2 + 1 + random.uniform(-0.2, 0.2)

    # Define a one-dimensional minimization problem with one real variable bounded by -2 and 3
    problem_config = ProblemConfig()
    problem_config.add_feature("real", (-2, 3))
    problem_config.add_min_objective()

    # Create training data using the randomly disturbed function f(x) = x^2 + 1 + eps
    X_train = np.reshape(np.linspace(-2, 3, 10), (-1, 1))
    y_train = np.reshape([my_func(x) for x in X_train], (-1, 1))

    # Define enting object and corresponding parameters
    params = EntingParams(unc_params=UncParams(
        dist_metric="l1",
        acq_sense="exploration"
    ))
    enting = Enting(problem_config, params=params)
    # Fit tree model
    enting.fit(X_train, y_train)

    # Build PyomoOptimizer object with Gurobi as solver and solve optimization problem
    params_pyo = {"solver_name": "gurobi"}
    opt_pyo = PyomoOptimizer(problem_config, params=params_pyo)
    res = opt_pyo.solve(enting)

    assert round(res.opt_point[0]) == 0


def test_compare_pyomo_gurobipy_multiobj():
    """
    Ensures for a multi objective example with l1  and l2 uncertainty metric and mixed feature types that optimization
    results for Gurobipy model and Pyomo model with Gurobi as optimizer coincide.
    """

    # define problem
    problem_config = ProblemConfig(rnd_seed=73)
    # number of objectives
    number_objectives = 2
    build_multi_obj_categorical_problem(problem_config, n_obj=number_objectives)

    # sample data
    rnd_sample = problem_config.get_rnd_sample_list(num_samples=20)
    testfunc_evals = eval_multi_obj_cat_testfunc(rnd_sample, n_obj=number_objectives)

    for metric in ["l1", "l2", "euclidean_squared"]:
        for acq_sense in ["exploration", "penalty"]:
            params = EntingParams(unc_params=UncParams(
                dist_metric=metric,
                acq_sense=acq_sense
            ))
            enting = Enting(problem_config, params=params)
            # fit tree ensemble
            enting.fit(rnd_sample, testfunc_evals)

            # Build GurobiOptimizer object and solve optimization problem
            params_gurobi = {"NonConvex": 2, "MIPGap": 0}
            opt_gur = GurobiOptimizer(problem_config, params=params_gurobi)
            res_gur = opt_gur.solve(enting, weights=(0.4, 0.6))

            # Build PyomoOptimizer object with Gurobi as solver and solve optimization problem
            params_pyo = {
                "solver_name": "gurobi",
                "solver_options": {"NonConvex": 2, "MIPGap": 0},
            }
            opt_pyo = PyomoOptimizer(problem_config, params=params_pyo)
            res_pyo = opt_pyo.solve(enting, weights=(0.4, 0.6))

            # Assert that active leaves coincide for both models
            assert math.isclose(res_gur.opt_val, res_pyo.opt_val, abs_tol=0.01)


# @pytest.mark.pipeline_test
def test_compare_pyomo_gurobipy_singleobj():
    """
    Ensures for a single objective example with l1  and l2 uncertainty metric and mixed feature types that optimization
    results for Gurobipy model and Pyomo model with Gurobi as optimizer coincide.
    """

    # define problem
    problem_config = ProblemConfig(rnd_seed=73)
    # number of objectives
    number_objectives = 1
    build_multi_obj_categorical_problem(problem_config, n_obj=number_objectives)

    # sample data
    rnd_sample = problem_config.get_rnd_sample_list(num_samples=20)
    testfunc_evals = eval_multi_obj_cat_testfunc(rnd_sample, n_obj=number_objectives)

    for metric in ["l1", "l2", "euclidean_squared"]:
        for acq_sense in ["exploration", "penalty"]:
            params = EntingParams(unc_params=UncParams(
                dist_metric=metric,
                acq_sense=acq_sense
            ))            
            enting = Enting(problem_config, params=params)
            # fit tree ensemble
            enting.fit(rnd_sample, testfunc_evals)

            # Build GurobiOptimizer object and solve optimization problem
            params_gurobi = {"NonConvex": 2, "MIPGap": 1e-3}
            opt_gur = GurobiOptimizer(problem_config, params=params_gurobi)
            res_gur = opt_gur.solve(enting)

            # Build PyomoOptimizer object with Gurobi as solver and solve optimization problem
            params_pyo = {
                "solver_name": "gurobi",
                "solver_options": {"NonConvex": 2, "MIPGap": 1e-3},
            }
            opt_pyo = PyomoOptimizer(problem_config, params=params_pyo)
            res_pyo = opt_pyo.solve(enting)

            # Compare optimal values (e.g. objective values) ...
            assert round(res_gur.opt_val / res_pyo.opt_val, 5) <= 1.001
            # ... and optimal points (e.g. feature variables)
            assert [round(x) for x in res_gur.opt_point[2:]] == [
                round(x) for x in res_pyo.opt_point[2:]
            ]



from entmoot.models.uncertainty_models.base_distance import CatDistance
import numpy as np
from math import log


class OfDistance(CatDistance):
    def _get_of_frac(self, cat_rows, cat_left, cat_right):
        count_cat_left = np.sum(cat_rows == cat_left)
        count_cat_right = np.sum(cat_rows == cat_right)
        n_rows = len(cat_rows)
        return 1 / (1 + log(n_rows / count_cat_left) * log(n_rows / count_cat_right))

    def _sim_mat_rule(self, x_left, x_right, cat_idx):
        return (
            self._get_of_frac(self._cache_x[:, cat_idx], x_left, x_right)
            if x_left != x_right
            else 1.0
        )



from entmoot.models.uncertainty_models.base_distance import CatDistance
import numpy as np


class Goodall4Distance(CatDistance):
    def _get_pk2(self, cat_rows, cat):
        count_cat = np.sum(cat_rows == cat)
        n_rows = len(cat_rows)
        return (count_cat * (count_cat - 1)) / (n_rows * (n_rows - 1))

    def _sim_mat_rule(self, x_left, x_right, cat_idx):
        return (
            self._get_pk2(self._cache_x[:, cat_idx], x_left)
            if x_left == x_right
            else 0.0
        )



def read_lgbm_tree_model_dict(tree_model_dict, cat_idx):
    ordered_tree_list = []

    for tree in tree_model_dict["tree_info"]:

        # generate list of nodes in tree
        root_node = [tree["tree_structure"]]
        node_list = []

        # populate node_list and add to ordered_tree_list if non-empty
        add_next_nodes(node_list=node_list, node=root_node, cat_idx=cat_idx)

        if node_list:
            ordered_tree_list.append(node_list)

    return ordered_tree_list


def add_next_nodes(node_list, node, cat_idx):
    if node:
        # add new node definition
        new_node = {}

        try:
            new_node["split_var"] = node[-1]["split_feature"]
            if not new_node["split_var"] in cat_idx:
                # read numerical variables, solver accuracy 10e-5
                temp_node_val = round(node[-1]["threshold"], 5)
                new_node["split_code_pred"] = temp_node_val
            else:
                # read categorical variables
                cat_set = node[-1]["threshold"].split("||")
                temp_node_val = [int(cat) for cat in cat_set]
                new_node["split_code_pred"] = temp_node_val

        except KeyError:
            # arrived at leaf node
            new_node["split_var"] = -1
            temp_node_val = node[-1]["leaf_value"]
            new_node["split_code_pred"] = temp_node_val

        node_list.append(new_node)

        # move on to next node in tree
        temp_node = node.pop(-1)
        try:
            node.append(temp_node["right_child"])
        except KeyError:
            pass

        try:
            node.append(temp_node["left_child"])
        except KeyError:
            pass

        add_next_nodes(node_list, node, cat_idx=cat_idx)



class MetaTreeModel:
    def __init__(self, tree_list: list):
        self._num_trees = len(tree_list)
        self.load_ordered_tree_list(tree_list)

    @property
    def num_trees(self):
        return self._num_trees

    def _build_tree(self, tree):
        node = tree.pop(0)
        split_var = node["split_var"]
        split_code_pred = node["split_code_pred"]

        return TreeNode(split_var=split_var, split_code_pred=split_code_pred, tree=tree)

    def load_ordered_tree_list(self, tree_list: list):
        self.trees = [self._build_tree(tree) for tree in tree_list]

    def get_leaf_encodings(self, tree):
        yield from self.trees[tree].get_leaf_encodings()

    def get_leaf_weight(self, tree, encoding):
        return self.trees[tree].get_leaf_weight(encoding)

    def get_var_break_points(self):
        var_breakpoints = {}
        for tree in self.trees:
            for var, breakpoint in tree.get_all_partition_pairs():
                try:
                    if isinstance(breakpoint, list):
                        var_breakpoints[var].append(breakpoint)
                    else:
                        var_breakpoints[var].add(breakpoint)
                except KeyError:
                    if isinstance(breakpoint, list):
                        var_breakpoints[var] = [breakpoint]
                    else:
                        var_breakpoints[var] = set([breakpoint])

        for k in var_breakpoints:
            if isinstance(var_breakpoints[k], set):
                var_breakpoints[k] = sorted(var_breakpoints[k])

        return var_breakpoints

    def get_participating_variables(self, tree, leaf):
        return set(self.trees[tree].get_participating_variables(leaf))

    def get_branch_encodings(self, tree):
        yield from self.trees[tree].get_branch_encodings()

    def get_branch_partition_pair(self, tree, encoding):
        return self.trees[tree].get_branch_partition_pair(encoding)

    def get_left_leaves(self, tree, encoding):
        yield from (encoding + s for s in self.trees[tree].get_left_leaves(encoding))

    def get_right_leaves(self, tree, encoding):
        yield from (encoding + s for s in self.trees[tree].get_right_leaves(encoding))

    def prune_var_bnds(self, encodings, var_bnds):
        # prunes var_bnds provided according to which splits are active
        for tree_id, leaf_enc in encodings:
            self.trees[tree_id]._prune_var_bnds(0, leaf_enc, var_bnds)
        return var_bnds


class TreeType:
    pass


class TreeNode(TreeType):
    def __init__(self, split_var, split_code_pred, tree):
        self.split_var = split_var
        self.split_code_pred = split_code_pred

        # check if tree is empty
        assert (
            tree
        ), "Given the data and train configuration, no tree ensemble could be trained."

        # read left node
        node = tree.pop(0)
        split_var = node["split_var"]
        split_code_pred = node["split_code_pred"]

        # split_var value of -1 refers to leaf node
        if split_var == -1:
            self.left = LeafNode(split_code_pred=split_code_pred)
        else:
            self.left = TreeNode(
                split_var=split_var, split_code_pred=split_code_pred, tree=tree
            )

        # read right node
        node = tree.pop(0)
        split_var = node["split_var"]
        split_code_pred = node["split_code_pred"]

        if split_var == -1:
            self.right = LeafNode(split_code_pred=split_code_pred)
        else:
            self.right = TreeNode(
                split_var=split_var, split_code_pred=split_code_pred, tree=tree
            )

    def __repr__(self):
        return ", ".join([str(x) for x in [self.split_var, self.split_code_pred]])

    def get_leaf_encodings(self, current_string=""):
        yield from self.left.get_leaf_encodings(current_string + "0")
        yield from self.right.get_leaf_encodings(current_string + "1")

    def get_leaf_weight(self, encoding):
        next_node = self.right if int(encoding[0]) else self.left
        return next_node.get_leaf_weight(encoding[1:])

    def get_all_partition_pairs(self):
        yield (self.split_var, self.split_code_pred)
        yield from self.left.get_all_partition_pairs()
        yield from self.right.get_all_partition_pairs()

    def get_participating_variables(self, encoding):
        yield self.split_var
        next_node = self.right if int(encoding[0]) else self.left
        yield from next_node.get_participating_variables(encoding[1:])

    def get_branch_encodings(self, current_string=""):
        yield current_string
        yield from self.left.get_branch_encodings(current_string + "0")
        yield from self.right.get_branch_encodings(current_string + "1")

    def _get_next_node(self, direction):
        return self.right if int(direction) else self.left

    def get_branch_partition_pair(self, encoding):
        if not encoding:
            return self.split_var, self.split_code_pred
        else:
            next_node = self._get_next_node(encoding[0])
            return next_node.get_branch_partition_pair(encoding[1:])

    def get_left_leaves(self, encoding):
        if encoding:
            next_node = self._get_next_node(encoding[0])
            yield from next_node.get_left_leaves(encoding[1:])
        else:
            yield from self.left.get_leaf_encodings("0")

    def get_right_leaves(self, encoding):
        if encoding:
            next_node = self._get_next_node(encoding[0])
            yield from next_node.get_right_leaves(encoding[1:])
        else:
            yield from self.right.get_leaf_encodings("1")

    def _prune_var_bnds(self, curr_depth, leaf_enc, var_bnds):
        if self.split_var != -1:
            if isinstance(self.split_code_pred, list):
                # categorical variable
                cat_set = set(self.split_code_pred)
                if leaf_enc[curr_depth] == "0":
                    var_bnds[self.split_var] = set(
                        var_bnds[self.split_var]
                    ).intersection(cat_set)
                    self.left._prune_var_bnds(curr_depth + 1, leaf_enc, var_bnds)
                else:
                    var_bnds[self.split_var] = set(var_bnds[self.split_var]).difference(
                        cat_set
                    )
                    self.right._prune_var_bnds(curr_depth + 1, leaf_enc, var_bnds)
            else:
                # continuous variable
                lb, ub = var_bnds[self.split_var]
                if leaf_enc[curr_depth] == "0":
                    ub = min(ub, self.split_code_pred)
                    var_bnds[self.split_var] = (lb, ub)
                    self.left._prune_var_bnds(curr_depth + 1, leaf_enc, var_bnds)
                else:  # if value is '1'
                    lb = max(lb, self.split_code_pred)
                    var_bnds[self.split_var] = (lb, ub)
                    self.right._prune_var_bnds(curr_depth + 1, leaf_enc, var_bnds)


class LeafNode(TreeNode):
    def __init__(self, split_code_pred):
        self.split_var = -1
        self.split_code_pred = split_code_pred

    def __repr__(self):
        return ", ".join([str(x) for x in ["Leaf", self.split_code_pred]])

    def get_leaf_encodings(self, current_string=""):
        yield current_string

    def get_leaf_weight(self, encoding):
        assert not encoding
        return self.split_code_pred

    def get_all_partition_pairs(self):
        yield from []

    def get_participating_variables(self, encoding):
        assert not encoding
        yield from []

    def get_branch_encodings(self, current_string=""):
        yield from []



from typing import TYPE_CHECKING, Callable
from abc import ABC, abstractmethod

import pyomo.environ as pyo

from entmoot.problem_config import FeatureType

if TYPE_CHECKING:
    from problem_config import FeatureType

ConstraintFunctionType = Callable[[pyo.ConcreteModel, int], pyo.Expression]


class Constraint(ABC):
    """A constraint to be applied to a model.

    Implements a user-friendly way to construct constraints to an optimisation problem.

    Attributes:
        feature_keys: A list of the string names of the features to be constrained"""

    def __init__(self, feature_keys: list[str]):
        self.feature_keys = feature_keys

    def _get_feature_vars(
        self, model: pyo.ConcreteModel, feat_list: list["FeatureType"]
    ) -> list[pyo.Var]:
        """Return a list of all the pyo.Vars, in the order of the constraint definition"""
        all_keys = [feat.name for feat in feat_list]
        feat_idxs = [all_keys.index(key) for key in self.feature_keys]
        features = [model._all_feat[i] for i in feat_idxs]
        return features

    @abstractmethod
    def as_pyomo_constraint(
        self, model: pyo.ConcreteModel, feat_list: list["FeatureType"]
    ) -> pyo.Constraint:
        """Convert to a pyomo.Constraint object.

        This requires the model (to access the variables), and the feat_list (to access the feature names)
        """
        pass


class ConstraintList:
    """Contains multiple constraints to be applied at once."""

    def __init__(self, constraints: list[Constraint]):
        self._constraints = constraints

    def add(self, constraint: Constraint):
        self._constraints.append(constraint)

    def apply_pyomo_constraints(
        self,
        model: pyo.ConcreteModel,
        feat_list: list[FeatureType],
        pyo_constraint_list: pyo.ConstraintList,
    ) -> None:
        """Add constraints to a pyo.ConstraintList object.

        Requires creation of the pyo.ConstraintList outside of this class,
        to the user to specify the constraints name."""

        for constraint in self._constraints:
            features = constraint._get_feature_vars(model, feat_list)
            if not isinstance(constraint, ExpressionConstraint):
                raise TypeError("Only ExpressionConstraints are supported in a constraint list")

            expr = constraint._get_expr(model, features)
            pyo_constraint_list.add(expr)


class ExpressionConstraint(Constraint):
    """Constraints defined by pyomo.Expressions.

    For constraints that can be simply defined by an expression of variables.
    """

    def as_pyomo_constraint(
        self, model: pyo.ConcreteModel, feat_list: list["FeatureType"]
    ) -> pyo.Constraint:
        features = self._get_feature_vars(model, feat_list)
        return pyo.Constraint(expr=self._get_expr(model, features))

    @abstractmethod
    def _get_expr(self, model, features) -> pyo.Expression:
        pass


class FunctionalConstraint(Constraint):
    """A constraint that uses a functional approach.

    For constraints that require creating intermediate variables and access to the model.
    """

    def as_pyomo_constraint(
        self, model: pyo.ConcreteModel, feat_list: list["FeatureType"]
    ) -> pyo.Constraint:
        features = self._get_feature_vars(model, feat_list)
        return pyo.Constraint(rule=self._get_function(model, features))

    @abstractmethod
    def _get_function(
        self, model: pyo.ConcreteModel, features: list["FeatureType"]
    ) -> ConstraintFunctionType:
        pass


class LinearConstraint(ExpressionConstraint):
    """Constraint that is a function of X @ C, where X is the feature list, and C
    is the list of coefficients."""

    def __init__(self, feature_keys: list[str], coefficients: list[float], rhs: float):
        self.coefficients = coefficients
        self.rhs = rhs
        super().__init__(feature_keys)

    def _get_lhs(self, features: pyo.ConcreteModel) -> pyo.Expression:
        """Get the left-hand side of the linear constraint"""
        return sum(f * c for f, c in zip(features, self.coefficients))


class LinearEqualityConstraint(LinearConstraint):
    def _get_expr(self, model, features):
        return self._get_lhs(features) == self.rhs


class LinearInequalityConstraint(LinearConstraint):
    def _get_expr(self, model, features):
        return self._get_lhs(features) <= self.rhs


class NChooseKConstraint(ExpressionConstraint):
    """Constrain the number of active features to be bounded by min_count and max_count."""

    tol: float = 1e-6
    M: float = 1e6

    def __init__(
        self,
        feature_keys: list[str],
        min_count: int,
        max_count: int,
        none_also_valid: bool = False,
    ):
        self.min_count = min_count
        self.max_count = max_count
        self.none_also_valid = none_also_valid
        super().__init__(feature_keys)

    def _get_expr(self, model, features):
        # constrain the features using the binary variable y
        # where y indicates whether the feature is selected
        # y * tol <= x <= y * M
        # tol is sufficiently small, M is sufficiently large
        model.feat_selected = pyo.Var(
            range(len(features)), domain=pyo.Binary, initialize=0
        )
        model.ub_selected = pyo.ConstraintList()
        model.lb_selected = pyo.ConstraintList()

        for i in range(len(features)):
            model.ub_selected.add(expr=model.feat_selected[i] * self.M >= features[i])
            model.lb_selected.add(expr=model.feat_selected[i] * self.tol <= features[i])

        return sum(model.feat_selected.values()) <= self.max_count



"""Dataclasses containing the parameters for Enting models"""

from typing import Literal
from dataclasses import dataclass, field

class ParamValidationError(ValueError):
    """A model parameter takes an invalid value."""
    pass

@dataclass
class UncParams:
    """
    This dataclass contains all uncertainty parameters.
    """
    #: weight for penalty/exploration part in objective function
    beta: float = 1.96
    #: the predictions of the GBT model are cut off, if their absolute value exceeds
    #: bound_coeff * variance of the y-values.
    bound_coeff: float = 0.5
    #: "exploration": try to find good points far away from known training data,
    #: "penalty": stay close to explored areas and try to find even better points there.
    acq_sense: Literal["exploration", "penalty"] = "exploration"
    #: controls two different types of transformations by scaling/shifting.
    #: "normal": shift by lower bound, scale by difference of smalles and largest value
    #: "standard": shift by mean, scale by standard deviation
    dist_trafo: Literal["normal", "standard"] = "normal"
    #: compute distance measure using the l_1, the l_2 or the squared l_2 norm.
    dist_metric: Literal["euclidean_squared", "l1", "l2"] = "euclidean_squared"
    #: different ways to compute the distance of categorical features
    cat_metric: Literal["overlap", "of", "goodall4"] = "overlap"

    def __post_init__(self):
        if self.beta < 0.0:
            raise ParamValidationError(
                f"Value for 'beta' is {self.beta}; must be positive."
            )
        
        if self.acq_sense not in ("exploration", "penalty"):
            raise ParamValidationError(
                f"Value for 'acq_sense' is '{self.acq_sense}'; must be in ('exploration', 'penalty')."
            )
        

@dataclass
class TrainParams:
    """
    This dataclass contains all hyperparameters that are used by lightbm during training and
    documented here https://lightgbm.readthedocs.io/en/latest/Parameters.html
    """
    # lightgbm training hyperparameters
    objective: str = "regression"
    metric: str = "rmse"
    boosting: str = "gbdt"
    num_boost_round: int = 100
    max_depth: int = 3
    min_data_in_leaf: int = 1
    min_data_per_group: int = 1
    verbose: int = -1


@dataclass
class TreeTrainParams:
    """
    This dataclass contains all parameters needed for the tree training.
    """
    train_params: "TrainParams" = field(default_factory=dict)
    train_lib: Literal["lgbm"] = "lgbm"

    def __post_init__(self):
        if isinstance(self.train_params, dict):
            self.train_params = TrainParams(**self.train_params)
        
        if self.train_lib not in ("lgbm",):
            raise ParamValidationError(
                f"Value for 'train_lib' is {self.train_lib}; must be in ('lgbm',)"
            )


@dataclass
class EntingParams:
    """Contains parameters for a mean and uncertainty model.
    
    Provides a structured dataclass for the parameters of an Enting model, 
    alongside default values and some light data validation."""
    unc_params: "UncParams" = field(default_factory=dict)
    tree_train_params: "TreeTrainParams" = field(default_factory=dict)
    
    def __post_init__(self):
        if isinstance(self.unc_params, dict):
            self.unc_params = UncParams(**self.unc_params)

        if isinstance(self.tree_train_params, dict):
            self.tree_train_params = TreeTrainParams(**self.tree_train_params)


from typing import Tuple, List
import numpy as np
import random


class ProblemConfig:
    def __init__(self, rnd_seed: int = None):
        self._feat_list = []
        self._obj_list = []
        self._rnd_seed = rnd_seed

        if rnd_seed is not None:
            assert isinstance(
                rnd_seed, int
            ), f"Argument 'rnd_seed' needs to be an integer. Got type '{type(rnd_seed)}'."
            np.random.seed(rnd_seed)
            random.seed(rnd_seed)

    @property
    def cat_idx(self):
        return tuple([i for i, feat in enumerate(self.feat_list) if feat.is_cat()])

    @property
    def non_cat_idx(self):
        return tuple([i for i, feat in enumerate(self.feat_list) if not feat.is_cat()])

    @property
    def non_cat_lb(self):
        return tuple(
            [feat.lb for i, feat in enumerate(self.feat_list) if not feat.is_cat()]
        )

    @property
    def non_cat_ub(self):
        return tuple(
            [feat.ub for i, feat in enumerate(self.feat_list) if not feat.is_cat()]
        )

    @property
    def non_cat_bnd_diff(self):
        return tuple(
            [
                feat.ub - feat.lb
                for i, feat in enumerate(self.feat_list)
                if not feat.is_cat()
            ]
        )

    @property
    def feat_list(self):
        return self._feat_list

    @property
    def obj_list(self):
        return self._obj_list

    @property
    def rnd_seed(self):
        return self._rnd_seed

    def get_enc_bnd(self):
        return [feat.get_enc_bnds() for feat in self._feat_list]

    def _encode_xi(self, xi: List):
        return np.asarray(
            [feat.encode(xi[idx]) for idx, feat in enumerate(self.feat_list)]
        )

    def _decode_xi(self, xi: List):
        return [feat.decode(xi[idx]) for idx, feat in enumerate(self.feat_list)]

    def encode(self, X: List):
        if len(X) == 0:
            return []
        elif len(X) == 1:
            return self._encode_xi(X[0])
        else:
            enc = [self._encode_xi(xi) for xi in X]
            return np.asarray(enc)

    def decode(self, X: List):
        if len(X) == 0:
            return []
        elif len(X) == 1:
            return self._decode_xi(X[0])
        else:
            dec = [self._decode_xi(xi) for xi in X]
            return np.asarray(dec)

    def add_feature(self, feat_type: str, bounds: Tuple = None, name: str = None):
        if name is None:
            name = f"feat_{len(self.feat_list)}"

        if bounds is None and feat_type in ("real", "integer", "categorical"):
            raise IOError(
                "Please provide bounds for feature types in '(real, integer, categorical)'"
            )

        # perform basic input checks and define features of space
        if feat_type in ("real", "integer"):
            assert len(bounds) == 2, (
                f"Too many values for bounds for feat_type '{feat_type}'. Expected '2', "
                f"got '{len(bounds)}'. Check feature '{name}'"
            )

            lb, ub = bounds

            if feat_type == "real":
                if isinstance(lb, int):
                    lb = float(lb)
                if isinstance(ub, int):
                    ub = float(ub)

                assert isinstance(lb, float) and isinstance(ub, float), (
                    f"Wrong type for bounds of feat_type '{feat_type}'. Expected '(float, "
                    f"float)', got '({type(lb)}, {type(ub)})'. Check feature '{name}'"
                )

                assert lb < ub, (
                    f"Lower bound '{lb}' of feat_type '{feat_type}' needs to be "
                    f"smaller than upper bound. Check feature '{name}'."
                )

                self._feat_list.append(Real(lb=lb, ub=ub, name=name))
            else:
                assert isinstance(lb, int) and isinstance(ub, int), (
                    f"Wrong type for bounds of feat_type '{feat_type}'. Expected '(int, int)', "
                    f"got '({type(lb)}, {type(ub)})'. Check feature '{name}'"
                )

                assert lb < ub, (
                    f"Lower bound '{lb}' of feat_type '{feat_type}' needs to be "
                    f"smaller than upper bound. Check feature '{name}'."
                )

                self._feat_list.append(Integer(lb=lb, ub=ub, name=name))

        elif feat_type == "binary":
            self._feat_list.append(Binary(name=name))

        elif feat_type == "categorical":
            assert len(bounds) > 1, (
                f"Not enough categories specified for bounds of feat_type '{feat_type}'. "
                f"Expected 'len(bounds) > 1', got 'len(bounds) = {len(bounds)}'. "
                f"Check feature '{name}'."
            )

            assert all(isinstance(item, str) for item in bounds), (
                f"Wrong type for bounds of feat_type '{feat_type}'. Expected 'str' for all "
                f"categories."
            )

            assert len(bounds) == len(
                set(bounds)
            ), f"Categories of feat_type '{feat_type}' are not all unique."

            self._feat_list.append(Categorical(cat_list=bounds, name=name))

        else:
            raise IOError(
                f"No support for feat_type '{feat_type}'. Check feature '{name}'."
            )

    def add_min_objective(self, name: str = None):
        if name is None:
            name = f"obj_{len(self.obj_list)}"

        self._obj_list.append(MinObjective(name=name))

    def add_max_objective(self, name: str = None):
        if name is None:
            name = f"obj_{len(self.obj_list)}"

        self._obj_list.append(MaxObjective(name=name))

    def transform_objective(self, y: np.ndarray) -> np.ndarray:
        """Transform data for minimisation/maximisation"""
        # y.shape = (num_samples, num_obj)
        signs = np.array([obj.sign for obj in self.obj_list]).reshape(1, -1)
        return y * signs



    def get_rnd_sample_numpy(self, num_samples):
        # returns np.array for faster processing
        array_list = []
        for feat in self.feat_list:
            if feat.is_real():
                array_list.append(
                    np.random.uniform(low=feat.lb, high=feat.ub, size=num_samples)
                )
            elif feat.is_cat():
                array_list.append(
                    np.random.randint(len(feat.cat_list), size=num_samples)
                )
            elif feat.is_int() or feat.is_bin():
                array_list.append(
                    np.random.random_integers(
                        low=feat.lb, high=feat.ub, size=num_samples
                    )
                )
        return np.squeeze(np.column_stack(array_list))

    def get_rnd_sample_list(self, num_samples=1, cat_enc=False):
        # returns list of tuples
        sample_list = []
        for _ in range(num_samples):
            sample = []
            for feat in self.feat_list:
                if feat.is_real():
                    sample.append(random.uniform(feat.lb, feat.ub))
                elif feat.is_cat():
                    if cat_enc:
                        sample.append(random.randrange(0, len(feat.cat_list)))
                    else:
                        sample.append(random.choice(feat.cat_list))
                elif feat.is_int() or feat.is_bin():
                    sample.append(random.randint(feat.lb, feat.ub))
            sample_list.append(tuple(sample))
        return sample_list if len(sample_list) > 1 else sample_list[0]

    def get_gurobi_model_core(self, env=None):
        import gurobipy as gur

        # initialize gurobi model
        if env is None:
            model = gur.Model()
        else:
            model = gur.Model(env=env)

        # initalize var space
        model._all_feat = []

        for i, feat in enumerate(self.feat_list):
            if feat.is_real():
                model._all_feat.append(
                    model.addVar(lb=feat.lb, ub=feat.ub, name=feat.name, vtype="C")
                )
            elif feat.is_cat():
                model._all_feat.append(dict())
                for enc, cat in zip(feat.enc_cat_list, feat.cat_list):
                    comb_name = f"{feat.name}_{cat}"
                    model._all_feat[i][enc] = model.addVar(name=comb_name, vtype="B")
            elif feat.is_int():
                model._all_feat.append(
                    model.addVar(lb=feat.lb, ub=feat.ub, name=feat.name, vtype="I")
                )
            elif feat.is_bin():
                model._all_feat.append(model.addVar(name=feat.name, vtype="B"))

        model.update()
        return model

    def copy_gurobi_model_core(self, model_core):
        """
        Computes a copy of a Gurobi model which is decoupled from the original, i.e. you can modify the copy without
        changing the original model.
        :param model_core: The gurobi model we want to copy
        :type model_core: gurobi.Model
        :return:
        :rtype: gurobi.Model
        """
        copy_model_core = model_core.copy()
        # Since Gurobi's .copy() method does not copy attributes like ._all_feat we have to copy them manually
        copy_model_core._all_feat = []

        # transfer feature var list to model copy
        for i, feat in enumerate(self.feat_list):
            if feat.is_cat():
                copy_model_core._all_feat.append(dict())
                for enc, cat in zip(feat.enc_cat_list, feat.cat_list):
                    var_name = model_core._all_feat[i][enc].VarName
                    copy_model_core._all_feat[i][enc] = copy_model_core.getVarByName(
                        var_name
                    )
            else:
                var_name = model_core._all_feat[i].VarName
                copy_model_core._all_feat.append(copy_model_core.getVarByName(var_name))
        return copy_model_core

    def copy_pyomo_model_core(self, model_core):
        """
        Computes a copy of a Pyomo model which is decoupled from the original, i.e. you can modify the copy without
        changing the original model.
        :param model_core: A Pyomo model
        :type model_core: pyomo.environ.ConcreteModel
        :return: A copy of the Pyomo model
        :rtype: pyomo.environ.ConcreteModel
        """
        return model_core.clone()

    def get_pyomo_model_core(self):
        import pyomo.environ as pyo

        # initialize Pyomo model
        model = pyo.ConcreteModel()

        # Initialize var space
        model._all_feat = []

        # Define some auxiliary dictionaries
        index_to_var_domain = {}
        index_to_var_bounds = {}
        for i, feat in enumerate(self.feat_list):
            if feat.is_real():
                index_to_var_domain[i] = pyo.Reals
                index_to_var_bounds[i] = (feat.lb, feat.ub)
            elif feat.is_cat():
                for enc, cat in zip(feat.enc_cat_list, feat.cat_list):
                    # We encode the index of this variable by (i, enc, cat), where 'i' is the position in the list of
                    # features, 'enc' is the corresponding encoded numerical value and 'cat' is the category
                    index_to_var_domain[i, enc, cat] = pyo.Binary
            elif feat.is_int():
                index_to_var_domain[i] = pyo.Integers
                index_to_var_bounds[i] = (feat.lb, feat.ub)
            elif feat.is_bin():
                index_to_var_domain[i] = pyo.Binary

        # Build Pyomo index set from dictionary keys
        # Why strings?? Well, pyomo doesn't support lists where some indices are multidimensional, but this is handy
        # for encoded categorical features
        model.indices_features = pyo.Set(
            initialize=[str(x) for x in index_to_var_domain]
        )

        # These auxiliary functions are needed, because Pyomo does not accept the dictionaries instead. The 'model'
        # argument is needed as well, but not explicitly used.
        # We hve the eval-statement eval(i) instead of just i, because the indices are strings for reasons explained
        # above :(
        def i_to_dom(model, i):
            return index_to_var_domain[eval(i)]

        def i_to_bounds(model, i):
            if eval(i) in index_to_var_bounds:
                return index_to_var_bounds[eval(i)]
            else:
                return (None, None)

        # Define decision variables for Pyomo model
        model.x = pyo.Var(model.indices_features, domain=i_to_dom, bounds=i_to_bounds)
        # We store the decision variables corresponding to the features in the list _all_feat in order to follow a
        # similar strategy to the gurobi model. Can be replaced in the future by a more direct approach.
        for i_str in model.x:
            i = eval(i_str)
            if type(i) is int:
                model._all_feat.append(model.x[i_str])
            elif type(i) is tuple:
                if i[1] == 0:
                    model._all_feat.append({})
                    model._all_feat[-1][i[1]] = model.x[i_str]
                elif i[1] > 0:
                    model._all_feat[-1][i[1]] = model.x[i_str]
                else:
                    raise TypeError
            else:
                raise TypeError

        return model

    def __str__(self):
        out_str = list(["\nPROBLEM SUMMARY"])
        out_str.append(len(out_str[-1][:-1]) * "-")
        out_str.append("features:")
        for feat in self.feat_list:
            if feat.is_cat():
                out_str.append(
                    f"{feat.name} :: {feat.__class__.__name__} :: {feat.cat_list} "
                )
            else:
                out_str.append(
                    f"{feat.name} :: {feat.__class__.__name__} :: ({feat.lb}, {feat.ub}) "
                )

        out_str.append("\nobjectives:")
        for obj in self.obj_list:
            out_str.append(f"{obj.name} :: {obj.__class__.__name__}")
        return "\n".join(out_str)


class FeatureType:
    def get_enc_bnds(self):
        return (self.lb, self.ub)

    def is_real(self):
        return False

    def is_cat(self):
        return False

    def is_int(self):
        return False

    def is_bin(self):
        return False

    def encode(self, xi):
        return xi

    def decode(self, xi):
        return xi


class Real(FeatureType):
    def __init__(self, lb, ub, name):
        self.lb = lb
        self.ub = ub
        self.name = name

    def is_real(self):
        return True


class Categorical(FeatureType):
    def __init__(self, cat_list, name):
        self._cat_list = cat_list
        self.name = name

        # encode categories
        self._enc2str, self._str2enc = {}, {}
        self._enc_cat_list = []
        for enc, cat in enumerate(cat_list):
            self._enc_cat_list.append(enc)
            self._enc2str[enc] = cat
            self._str2enc[cat] = enc

    def get_enc_bnds(self):
        return self._enc_cat_list

    @property
    def cat_list(self):
        return self._cat_list

    @property
    def enc_cat_list(self):
        return self._enc_cat_list

    def encode(self, xi):
        return self._str2enc[xi]

    def decode(self, xi):
        return self._enc2str[xi]

    def is_cat(self):
        return True


class Integer(FeatureType):
    def __init__(self, lb, ub, name):
        self.lb = lb
        self.ub = ub
        self.name = name

    def is_int(self):
        return True

    def decode(self, xi):
        return int(xi)


class Binary(FeatureType):
    def __init__(self, name):
        self.lb = 0
        self.ub = 1
        self.name = name

    def decode(self, xi):
        return abs(int(xi))

    def is_bin(self):
        return True

class Objective:
    def __init__(self, name):
        self.name = name

class MinObjective(Objective):
    sign = 1

class MaxObjective(Objective):
    sign = -1


from entmoot.models.model_params import EntingParams, UncParams, TrainParams, TreeTrainParams, ParamValidationError
import pytest

def test_model_params_creation():
    """Check EntingParams is instantiated correctly, and check default values."""
    params = EntingParams(**{
        "unc_params": {"beta": 2},
        "tree_train_params" : {
            "train_params": {"max_depth": 5}
            }
    })

    assert params.unc_params.beta == 2
    assert params.tree_train_params.train_params.max_depth == 5

    # check a selection of defaults
    assert params.unc_params.acq_sense in ("exploration", "penalty")
    assert params.tree_train_params.train_params.min_data_in_leaf == 1

    # check alternate initialisation method
    params_other = EntingParams(
        unc_params=UncParams(beta=2),
        tree_train_params=TreeTrainParams(
            train_params=TrainParams(max_depth=5)
        )
    )
    assert params == params_other


def test_model_params_invalid_values():
    """Check EntingParams raises an error for invalid values."""
    with pytest.raises(ParamValidationError):
        _ = EntingParams(**{"unc_params": {"beta": -1}})

    with pytest.raises(ParamValidationError):
        _ = EntingParams(**{"tree_train_params": {"train_lib": "notimplementedlib"}})

    with pytest.raises(ParamValidationError):
        _ = EntingParams(**{"unc_params": {"acq_sense": "notimplementedsense"}})


from entmoot import Enting, ProblemConfig, GurobiOptimizer, PyomoOptimizer
from entmoot.benchmarks import (
    build_multi_obj_categorical_problem,
    eval_multi_obj_cat_testfunc,
)


def test_bring_your_own_constraints():
    # define problem
    problem_config = ProblemConfig(rnd_seed=73)
    # number of objectives
    number_objectives = 2
    build_multi_obj_categorical_problem(problem_config, n_obj=number_objectives)

    # sample data
    rnd_sample = problem_config.get_rnd_sample_list(num_samples=20)
    testfunc_evals = eval_multi_obj_cat_testfunc(rnd_sample, n_obj=number_objectives)

    params = {"unc_params": {"dist_metric": "l1", "acq_sense": "exploration"}}
    enting = Enting(problem_config, params=params)
    # fit tree ensemble
    enting.fit(rnd_sample, testfunc_evals)

    # Gurobi Version

    # get optimization model
    model_gur = problem_config.get_gurobi_model_core()
    # extract decision variables
    x = model_gur._all_feat[3]
    y = model_gur._all_feat[4]
    z = model_gur._all_feat[5]
    # add constraint that all variables should coincide
    model_gur.addConstr(x == y)
    model_gur.addConstr(y == z)

    model_gur.update()

    # Build GurobiOptimizer object and solve optimization problem
    params_gurobi = {"MIPGap": 0}
    opt_gur = GurobiOptimizer(problem_config, params=params_gurobi)

    res_gur = opt_gur.solve(enting, model_core=model_gur)
    x_opt, y_opt, z_opt = res_gur.opt_point[3:]

    assert round(x_opt, 5) == round(y_opt, 5) and round(y_opt, 5) == round(z_opt, 5)


from entmoot.models.base_model import BaseModel


class TreeKernelEntmoot(BaseModel):
    def add_to_gurobipy_model(core_model, gurobi_env):
        return NotImplementedError()

    def add_to_pyomo_model(core_model):
        return NotImplementedError()



from collections import namedtuple
import numpy as np
from scipy.special import comb


OptResult = namedtuple(
    "OptResult",
    ["opt_point", "opt_val", "mu_unscaled", "unc_unscaled", "active_leaf_enc"],
)


def grid(dimension: int, levels: int) -> np.ndarray:
    """Construct a regular grid on the unit simplex.

    The number of grid points is L = (n+m-1)!/(n!*(m-1)!)
    where m is the dimension and n+1 is the number of levels.

    Args:
        dimension (int): Number of variables.
        levels (int): Number of levels for each variable.

    Returns:
       array: Regularily spaced points on the unit simplex.

    Examples:
        --> simplex_grid(3, 3)
        array([
            [0. , 0. , 1. ],
            [0. , 0.5, 0.5],
            [0. , 1. , 0. ],
            [0.5, 0. , 0.5],
            [0.5, 0.5, 0. ],
            [1. , 0. , 0. ]
        ])

    References:
        Nijenhuis and Wilf, Combinatorial Algorithms, Chapter 5, Academic Press, 1978.
    """
    m = dimension
    n = levels - 1
    L = int(comb(dimension - 1 + levels - 1, dimension - 1, exact=True))

    x = np.zeros(m, dtype=int)
    x[-1] = n

    out = np.empty((L, m), dtype=int)
    out[0] = x

    h = m
    for i in range(1, L):
        h -= 1

        val = x[h]
        x[h] = 0
        x[h - 1] += 1
        x[-1] = val - 1

        if val != 1:
            h = m

        out[i] = x

    return out / n


def sample(dimension: int, n_samples: int = 1) -> np.ndarray:
    """Sample uniformly from the unit simplex.

    Args:
        dimension (int): Number of dimensions.
        n_samples (int): Number of samples to draw.

    Returns:
        array, shape=(n_samples, dimesnion): Random samples from the unit simplex.
    """
    s = np.random.standard_exponential((n_samples, dimension))
    return (s.T / s.sum(axis=1)).T



from entmoot import Enting, ProblemConfig, PyomoOptimizer
from entmoot.benchmarks import (
    build_multi_obj_categorical_problem,
    eval_multi_obj_cat_testfunc,
)
from pytest import approx

def test_max_predictions_equal_min_predictions():
    """The sign of the predicted objective is independent of max/min."""
    problem_config = ProblemConfig(rnd_seed=73)
    build_multi_obj_categorical_problem(problem_config, n_obj=1)
    problem_config.add_min_objective()

    problem_config_max = ProblemConfig(rnd_seed=73)
    build_multi_obj_categorical_problem(problem_config_max, n_obj=1)
    problem_config_max.add_max_objective()

    rnd_sample = problem_config.get_rnd_sample_list(num_samples=20)
    testfunc_evals = eval_multi_obj_cat_testfunc(rnd_sample, n_obj=2)

    params = {"unc_params": {"dist_metric": "l1", "acq_sense": "exploration"}}
    enting = Enting(problem_config, params=params)
    enting.fit(rnd_sample, testfunc_evals)

    enting_max = Enting(problem_config_max, params=params)
    enting_max.fit(rnd_sample, testfunc_evals)

    sample = problem_config.get_rnd_sample_list(num_samples=3)
    pred = enting.predict(sample)
    pred_max = enting_max.predict(sample)

    for ((m1, u1), (m2, u2)) in zip(pred, pred_max):
        print(">", m1, m2)
        assert m1 == approx(m2, rel=1e-5)
        assert u1 == approx(u2, rel=1e-5)

def test_max_objective_equals_minus_min_objective():
    """Assert that the solution found by the minimiser is the same as that of the maximiser for the negative objective function"""
    problem_config = ProblemConfig(rnd_seed=73)
    build_multi_obj_categorical_problem(problem_config, n_obj=1)
    problem_config.add_min_objective()

    problem_config_max = ProblemConfig(rnd_seed=73)
    build_multi_obj_categorical_problem(problem_config_max, n_obj=0)
    problem_config_max.add_max_objective()
    problem_config_max.add_max_objective()

    rnd_sample = problem_config.get_rnd_sample_list(num_samples=20)
    testfunc_evals = eval_multi_obj_cat_testfunc(rnd_sample, n_obj=2)

    params = {"unc_params": {"dist_metric": "l1", "acq_sense": "penalty"}}
    enting = Enting(problem_config, params=params)
    enting.fit(rnd_sample, testfunc_evals)
    # pass negative test evaluations to the maximiser
    enting_max = Enting(problem_config, params=params)
    enting_max.fit(rnd_sample, -testfunc_evals)

    params_pyomo = {"solver_name": "gurobi"}
    res = PyomoOptimizer(problem_config, params=params_pyomo).solve(enting)
    res_max = PyomoOptimizer(problem_config_max, params=params_pyomo).solve(enting)

    assert res.opt_point == approx(res_max.opt_point, rel=1e-5)






from collections import namedtuple
from entmoot import Enting, ProblemConfig
from entmoot.utils import OptResult
import gurobipy as gur
import os


class GurobiOptimizer:
    """
    This class builds and solves a Gurobi optimization model using available
    information (tree structure, uncertainty measures, ...).

    Example:
        .. code-block:: python

            from entmoot import ProblemConfig
            import numpy as np
            import random

            # Define a one-dimensional minimization problem with one real variable bounded by -2 and 3
            problem_config = ProblemConfig()
            problem_config.add_feature("real", (-2, 3))
            problem_config.add_min_objective()

            # Create training data using the randomly disturbed function f(x) = x^2 + 1 + eps
            X_train = np.linspace(-2, 3, 10)
            y_train = [x**2 + 1 + random.uniform(-0.2, 0.2) for x in X_train]

            # Define enting object and corresponding parameters
            params = {"unc_params": {"dist_metric": "l1"}}
            enting = Enting(problem_config, params=params)
            # Fit tree model
            enting.fit(X_train, y_train)
            # Compute the predictions for training data and see that light gbm fitted a step function
            # with three steps
            enting.predict(X_train)

            # Define parameters needed during optimization
            params_gur = {"MIPGap": 1e-3}
            # Build GurobiOptimizer object.
            opt_gur = GurobiOptimizer(problem_config, params=params_gurobi)
            # As expected, the optimal input of the tree model is near the origin (cf. X_opt_pyo)
            X_opt_pyo, _, _ = opt_gur.solve(enting)
    """
    def __init__(self, problem_config: ProblemConfig, params: dict = None) -> float:
        self._params = {} if params is None else params
        self._problem_config = problem_config
        self._curr_sol = None
        self._active_leaves = None

    def get_curr_sol(self) -> list:
        """
        Returns current solution (i.e. optimal points) from optimization run
        """
        assert self._curr_sol is not None, "No solution was generated yet."
        return self._curr_sol

    def get_active_leaf_sol(self) -> list:
        """
        Returns active leaves in the tree model based on the current solution
        """
        assert self._curr_sol is not None, "No solution was generated yet."
        return self._active_leaves

    def solve(
        self,
        tree_model: Enting,
        model_core: gur.Model = None,
        weights: tuple = None,
        use_env: bool = False,
    ) -> OptResult:
        """
        Solves the Gurobi optimization model
        """

        if model_core is None:
            if use_env:
                if "CLOUDACCESSID" in os.environ:
                    # Use Gurobi Cloud
                    connection_params_cld = {
                        "CLOUDACCESSID": os.getenv("CLOUDACCESSID"),
                        "CLOUDSECRETKEY": os.getenv("CLOUDSECRETKEY"),
                        "CLOUDPOOL": os.getenv("CLOUDPOOL"),
                    }
                    env_cld = gur.Env(params=connection_params_cld)
                    env_cld.start()
                    opt_model = self._problem_config.get_gurobi_model_core(env=env_cld)

            else:
                opt_model = self._problem_config.get_gurobi_model_core()
        else:
            # create model copy to not overwrite original one
            opt_model = self._problem_config.copy_gurobi_model_core(model_core)

        # check weights
        if weights is not None:
            assert len(weights) == len(self._problem_config.obj_list), (
                f"Number of 'weights' is '{len(weights)}', number of objectives "
                f"is '{len(self._problem_config.obj_list)}'."
            )
            assert sum(weights) == 1.0, "weights don't add up to 1.0"

        # set solver parameters
        for param, param_val in self._params.items():
            opt_model.setParam(param, param_val)

        # build gurobi model
        tree_model.add_to_gurobipy_model(opt_model, weights=weights)

        # Solve optimization model
        opt_model.optimize()

        # update current solution
        self._curr_sol, self._active_leaves = self._get_sol(opt_model)

        return OptResult(
            self.get_curr_sol(),
            opt_model.ObjVal,
            [x.X for x in opt_model._unscaled_mu],
            opt_model._unc.X,
            self._active_leaves,
        )

    def _get_sol(self, solved_model: gur.Model) -> list:
        # extract solutions from conti and discrete variables
        res = []
        for idx, feat in enumerate(self._problem_config.feat_list):
            curr_var = solved_model._all_feat[idx]
            if feat.is_cat():
                # find active category
                sol_cat = [
                    int(round(curr_var[enc_cat].x)) for enc_cat in feat.enc_cat_list
                ].index(1)
                res.append(sol_cat)
            else:
                res.append(curr_var.x)

        # extract active leaves of solution
        def obj_leaf_index(model_obj, obj_name):
            # this function is the same as in 'tree_ensemble.py', TODO: put this in a tree_utils?
            for tree in range(model_obj._num_trees(obj_name)):
                for leaf in model_obj._leaves(obj_name, tree):
                    yield tree, leaf

        act_leaves = []
        for idx, obj in enumerate(self._problem_config.obj_list):
            act_leaves.append(
                [
                    (tree_id, leaf_enc)
                    for tree_id, leaf_enc in obj_leaf_index(solved_model, obj.name)
                    if round(solved_model._z[obj.name, tree_id, leaf_enc].x) == 1.0
                ]
            )

        return self._problem_config.decode([res]), act_leaves



from entmoot.models.uncertainty_models.base_distance import NonCatDistance
import numpy as np


class L2Distance(NonCatDistance):
    def _get_distance(self, x_left, x_right):
        if x_left.ndim == 1:
            dist = np.sqrt(np.sum((x_left - x_right) ** 2))
        else:
            dist = np.sqrt(np.sum((x_left - x_right) ** 2, axis=1))
        return dist

    def _array_predict(self, X):
        raise NotImplementedError()

    def get_gurobipy_model_constr_terms(self, model):
        from gurobipy import quicksum

        feat = model._all_feat

        constr_list = []
        for xi in self.x_trafo:
            constr = quicksum(
                (xi[i] - (feat[idx] - self.shift[i]) / self.scale[i])
                * (xi[i] - (feat[idx] - self.shift[i]) / self.scale[i])
                for i, idx in enumerate(self._problem_config.non_cat_idx)
            )
            constr_list.append(constr)
        return constr_list

    def get_pyomo_model_constr_terms(self, model):

        feat = model._all_feat

        constr_list = []
        for xi in self.x_trafo:
            constr = sum(
                (xi[i] - (feat[idx] - self.shift[i]) / self.scale[i])
                * (xi[i] - (feat[idx] - self.shift[i]) / self.scale[i])
                for i, idx in enumerate(self._problem_config.non_cat_idx)
            )
            constr_list.append(constr)
        return constr_list



from entmoot import ProblemConfig
from entmoot.models.base_model import BaseModel
from entmoot.models.mean_models.tree_ensemble import TreeEnsemble
from entmoot.models.uncertainty_models.distance_based_uncertainty import (
    DistanceBasedUncertainty,
)
from entmoot.models.model_params import EntingParams
from dataclasses import asdict
import numpy as np
from typing import Union


class Enting(BaseModel):
    """
    This class provides a living space for your tree model. You can fit your model, use it for predictions and
    provide information that are needed to build optimization models that incorporate the tree structure of your model.

    Example:
        .. code-block:: python

            from entmoot import ProblemConfig
            import numpy as np
            import random

            # Define a one-dimensional minimization problem with one real variable bounded by -2 and 3
            problem_config = ProblemConfig()
            problem_config.add_feature("real", (-2, 3))
            problem_config.add_min_objective()

            # Create training data using the randomly disturbed function f(x) = x^2 + 1 + eps
            X_train = np.linspace(-2, 3, 10)
            y_train = [x**2 + 1 + random.uniform(-0.2, 0.2) for x in X_train]

            # Define enting object and corresponding parameters
            params = {"unc_params": {"dist_metric": "l1"}}
            enting = Enting(problem_config, params=params)
            # Fit tree model
            enting.fit(X_train, y_train)
            # Compute the predictions for training data and see that light gbm fitted a step function
            # with three steps
            enting.predict(X_train)

            # Define parameters needed during optimization
            params_pyo = {"solver_name": "gurobi", "solver_options": {"MIPGap": 0}}
            # Build PyomoOptimizer object. This step will internally call the methods
            # add_to_pyomo_model() or add_to_gurobipy_model(), resp., depending on the choice
            # of your optimizer
            opt_pyo = PyomoOptimizer(problem_config, params=params_pyo)
            # As expected, the optimal input of the tree model is near the origin (cf. X_opt_pyo)
            X_opt_pyo, _, _ = opt_pyo.solve(enting)
    """

    def __init__(self, problem_config: ProblemConfig, params: Union[EntingParams, dict, None]):
        if params is None:
            params = {}
        if isinstance(params, dict):
            params = EntingParams(**params)

        self._problem_config = problem_config

        # initialize mean model
        self.mean_model = TreeEnsemble(
            problem_config=problem_config, params=params.tree_train_params
        )

        # initialize unc model
        unc_params = params.unc_params
        self._acq_sense = unc_params.acq_sense

        self._beta = unc_params.beta

        if self._acq_sense == "exploration":
            self._beta = -self._beta

        self.unc_model = DistanceBasedUncertainty(
            problem_config=problem_config, params=unc_params
        )

    def fit(self, X: np.ndarray, y: np.ndarray) -> None:
        """
        Performs the training of you tree model using training data and labels
        """
        # encode categorical features
        X = self._problem_config.encode(X)

        # check dims of X and y
        if X.ndim == 1:
            X = np.atleast_2d(X)

        assert X.shape[-1] == len(self._problem_config.feat_list), (
            "Argument 'X' has wrong dimensions. "
            f"Expected '(num_samples, {len(self._problem_config.feat_list)})', got '{X.shape}'."
        )

        if y.ndim == 1:
            y = np.atleast_2d(y)

        assert (y.shape[-1] == 2 and len(self._problem_config.obj_list) == 1) or (
            y.shape[-1] == len(self._problem_config.obj_list)
        ), (
            "Argument 'y' has wrong dimensions. "
            f"Expected '(num_samples, {len(self._problem_config.obj_list)})', got '{y.shape}'."
        )
        y = self._problem_config.transform_objective(y)
        self.mean_model.fit(X, y)
        self.unc_model.fit(X, y)

    def leaf_bnd_predict(self, obj_name, leaf_enc):
        bnds = self._problem_config.get_enc_bnd()
        return self.mean_model.meta_tree_dict[obj_name].prune_var_bnds(leaf_enc, bnds)

    def predict(self, X: np.ndarray, is_enc=False) -> list:
        """
        Computes prediction value of tree model for X.
        """
        # encode categorical features
        if not is_enc:
            X = self._problem_config.encode(X)

        # check dims of X
        if X.ndim == 1:
            X = np.atleast_2d(X)

        assert X.shape[-1] == len(self._problem_config.feat_list), (
            "Argument 'X' has wrong dimensions. "
            f"Expected '(num_samples, {len(self._problem_config.feat_list)})', got '{X.shape}'."
        )

        mean_pred = self.mean_model.predict(X) #.tolist()
        unc_pred = self.unc_model.predict(X)
        
        mean_pred = self._problem_config.transform_objective(mean_pred)
        mean_pred = mean_pred.tolist()

        comb_pred = [(mean, unc) for mean, unc in zip(mean_pred, unc_pred)]
        return comb_pred

    def predict_acq(self, X: np.ndarray, is_enc=False) -> list:
        """
        Predicts value of acquisition function (which contains not only the mean value but also the uncertainty)
        """
        acq_pred = []
        comb_pred = self.predict(X, is_enc=is_enc)
        for mean, unc in comb_pred:
            acq_pred.append(mean + self._beta * unc)
        return acq_pred

    def add_to_gurobipy_model(self, core_model, weights: tuple = None) -> None:
        """
        Enriches the core model by adding variables and constraints based on information
        from the tree model.
        """
        from gurobipy import GRB
        from entmoot.utils import sample

        # add uncertainty model part
        self.unc_model.add_to_gurobipy_model(core_model)

        core_model._mu = core_model.addVar(
            lb=-GRB.INFINITY, ub=GRB.INFINITY, name=f"mean_obj", vtype="C"
        )

        if len(self._problem_config.obj_list) == 1:
            # single objective case
            self.mean_model.add_to_gurobipy_model(core_model, add_mu_var=True)
            core_model.addConstr(core_model._mu == core_model._aux_mu[0])
        else:
            # multi-objective case
            self.mean_model.add_to_gurobipy_model(
                core_model, add_mu_var=True, normalize_mean=True
            )
            if weights is not None:
                moo_weights = weights
            else:
                moo_weights = sample(len(self._problem_config.obj_list), 1)[0]

            for idx, obj in enumerate(self._problem_config.obj_list):
                core_model.addConstr(
                    core_model._mu >= moo_weights[idx] * core_model._aux_mu[idx],
                    name=f"weighted_mean_obj_{idx}",
                )

        core_model.setObjective(core_model._mu + self._beta * core_model._unc)
        core_model.update()

    def add_to_pyomo_model(self, core_model, weights: tuple = None) -> None:
        """
        Enriches the core model by adding variables and constraints based on information
        from the tree model.
        """
        import pyomo.environ as pyo
        from entmoot.utils import sample

        # add uncertainty model part
        self.unc_model.add_to_pyomo_model(core_model)

        core_model._mu = pyo.Var(domain=pyo.Reals)

        if len(self._problem_config.obj_list) == 1:
            # single objective case
            self.mean_model.add_to_pyomo_model(core_model, add_mu_var=True)
            # Get objective name
            obj_name = self._problem_config.obj_list[0].name
            core_model.constraint_link_mu_auxmu = pyo.Constraint(
                expr=core_model._aux_mu[obj_name] == core_model._mu
            )
        else:
            # multi-objective case
            self.mean_model.add_to_pyomo_model(
                core_model, add_mu_var=True, normalize_mean=True
            )
            if weights is not None:
                moo_weights = weights
            else:
                moo_weights = sample(len(self._problem_config.obj_list), 1)[0]

            objectives_position_name = list(
                enumerate([obj.name for obj in self._problem_config.obj_list])
            )

            def constrs_weights_auxmus(model, pos, objname):
                return model._mu >= moo_weights[pos] * core_model._aux_mu[objname]

            core_model.constr_coupling_mu_auxmu = pyo.Constraint(
                objectives_position_name,
                rule=constrs_weights_auxmus,
            )

        # Define objective function
        core_model.obj = pyo.Objective(
            expr=core_model._mu + self._beta * core_model._unc, sense=pyo.minimize
        )



from setuptools import setup, find_packages
from pathlib import Path

project_root = Path(__file__).resolve().parent

about = {}
version_path = project_root / "entmoot" / "__version__.py"
with version_path.open() as f:
    exec(f.read(), about)

setup(
    name="entmoot",
    author=about["__author__"],
    author_email=about["__author_email__"],
    license=about["__license__"],
    version=about["__version__"],
    url="https://github.com/cog-imperial/entmoot",
    packages=find_packages(exclude=["tests", "docs"]),
    install_requires=[
        "numpy<=2.0.0",
        "lightgbm==4.0.0",
        "gurobipy",
        "pyomo"
    ],
    setup_requires=[
        "numpy<=2.0.0",
        "lightgbm==4.0.0",
        "gurobipy",
        "pyomo"
    ],
)



class BaseModel:
    def __init__(self, problem_config, num_obj, params):
        raise NotImplementedError()

    def predict(self, X):
        raise NotImplementedError()

    def fit(self, X, y):
        raise NotImplementedError()

    def add_to_gurobipy_model(self, model_core):
        raise NotImplementedError()

    def add_to_pyomo_model(self, model_core):
        raise NotImplementedError()



from entmoot.models.base_model import BaseModel
import numpy as np


class NonCatDistance(BaseModel):
    def __init__(self, problem_config, acq_sense, dist_trafo):
        self._problem_config = problem_config
        self._acq_sense = acq_sense
        self._dist_trafo = dist_trafo

        self._shift, self._scale = None, None
        self._x_trafo = None

    @property
    def shift(self):
        return self._shift

    @property
    def scale(self):
        return self._scale

    @property
    def x_trafo(self):
        assert (
            self._x_trafo is not None
        ), f"Uncertainty model needs fit function call before it can predict."
        return self._x_trafo

    def get_big_m(self):
        lb = self._trafo(self._problem_config.non_cat_lb)
        ub = self._trafo(self._problem_config.non_cat_ub)
        return self._get_distance(lb, ub)

    def _get_distance(self, x_left, x_right):
        raise NotImplementedError()

    def _trafo(self, non_cat_x):
        return (non_cat_x - self._shift) / self._scale

    def predict(self, xi):
        non_cat_x = self._trafo(np.atleast_2d(xi)[:, self._problem_config.non_cat_idx])
        return self._get_distance(self.x_trafo, non_cat_x)

    def _array_predict(self, X):
        raise NotImplementedError()

    def fit(self, X):
        non_cat_x = X[:, self._problem_config.non_cat_idx]

        # define shift and scalar values for non-cat feats
        if self._dist_trafo == "normal":
            self._shift = np.asarray(self._problem_config.non_cat_lb)
            self._scale = np.asarray(self._problem_config.non_cat_bnd_diff)
        elif self._dist_trafo == "standard":
            self._shift = np.mean(np.asarray(non_cat_x), axis=0)
            self._scale = np.std(np.asarray(non_cat_x), axis=0)
        else:
            raise IOError(
                "Parameter 'dist_trafo' for uncertainty model needs to be "
                "in '('normal', 'standard')'."
            )

        self._x_trafo = self._trafo(non_cat_x)

    def get_gurobipy_model_constr(self, model_core):
        raise NotImplementedError()

    def get_pyomo_model_constr(self, model_core):
        raise NotImplementedError()


class CatDistance(BaseModel):
    def __init__(self, problem_config, acq_sense):
        self._problem_config = problem_config
        self._acq_sense = acq_sense

        self._cat_x = None
        self._sim_map = None
        self._cache_x = None

    @property
    def cache_x(self):
        assert (
            self._cache_x is not None
        ), f"Uncertainty model needs fit function call before it can predict."
        return self._cache_x

    @property
    def sim_map(self):
        assert (
            self._sim_map is not None
        ), f"Uncertainty model needs fit function call before it can predict."
        return self._sim_map

    def get_big_m(self):
        # categorical contribution has maximum of one
        return len(self._problem_config.cat_idx)

    def predict(self, xi):
        # iterate through every evaluation row
        sim_vec = np.zeros(len(self.cache_x))

        for idx in self._problem_config.cat_idx:
            pred_cat = int(xi[idx])
            cached_cats = (np.rint(self.cache_x[:, idx])).astype(int)
            sim_vec += self.sim_map[idx][pred_cat, cached_cats]

        # compute distance based on similarity
        dist_vec = (-1) * sim_vec + len(self._problem_config.cat_idx)
        return dist_vec

    def _sim_mat_rule(self, x_left, x_right, cat_idx):
        raise NotImplementedError()

    def fit(self, X):
        self._cache_x = X

        # generate similarity matrix for all data points
        self._sim_map = {}

        for idx in self._problem_config.cat_idx:
            feat = self._problem_config.feat_list[idx]
            all_cats = feat.enc_cat_list

            # creates similarity entries for all categories of all categorical features
            mat = np.fromfunction(
                np.vectorize(self._sim_mat_rule, otypes=[float]),
                (len(all_cats), len(all_cats)),
                dtype=int,
                cat_idx=idx,
            )
            self._sim_map[idx] = mat

    def get_gurobipy_model_constr_terms(self, model):
        feat = model._all_feat
        constr_list = []
        for xi in self.cache_x:
            constr = 0

            # iterate through all categories to check which distances are active
            for idx in self._problem_config.cat_idx:
                for cat in self._problem_config.feat_list[idx].enc_cat_list:
                    sim = self.sim_map[idx][cat, int(xi[idx])]
                    constr += (1 - sim) * feat[idx][cat]

            constr_list.append(constr)
        return constr_list

    def get_pyomo_model_constr_terms(self, model):
        feat = model._all_feat
        constr_list = []
        for xi in self.cache_x:
            constr = 0

            # iterate through all categories to check which distances are active
            for idx in self._problem_config.cat_idx:
                for cat in self._problem_config.feat_list[idx].enc_cat_list:
                    sim = self.sim_map[idx][cat, int(xi[idx])]
                    constr += (1 - sim) * feat[idx][cat]

            constr_list.append(constr)
        return constr_list



from entmoot.models.base_model import BaseModel
from entmoot.models.mean_models.lgbm_utils import read_lgbm_tree_model_dict
from entmoot.models.mean_models.meta_tree_ensemble import MetaTreeModel
from entmoot.models.model_params import TreeTrainParams
import warnings
from typing import Union
from dataclasses import asdict
import numpy as np


class TreeEnsemble(BaseModel):
    def __init__(self, problem_config, params: Union[TreeTrainParams, dict, None] = None):
        if params is None:
            params = {}
        if isinstance(params, dict):
            params = TreeTrainParams(**params)

        self._problem_config = problem_config
        self._train_lib = params.train_lib
        self._rnd_seed = problem_config.rnd_seed

        self._train_params = asdict(params.train_params)
        if self._rnd_seed is not None:
            self._train_params["random_state"] = self._rnd_seed

        self._tree_dict = None
        self._meta_tree_dict = {}
        self._min_y, self._max_y = None, None

    @property
    def tree_dict(self):
        assert (
            self._tree_dict is not None
        ), "No tree model is trained yet. Call '.fit(X, y)' first."
        return self._tree_dict

    @property
    def meta_tree_dict(self):
        assert (
            len(self._meta_tree_dict) > 0
        ), "No tree model is trained yet. Call '.fit(X, y)' first."
        return self._meta_tree_dict

    @property
    def min_y(self):
        return self._min_y

    @property
    def max_y(self):
        return self._max_y

    def fit(self, X, y):
        # train tree models for every objective
        if self._tree_dict is None:
            self._tree_dict = {}

        # cache min/max vals for normalization
        self._min_y = np.min(y, axis=0)
        self._max_y = np.max(y, axis=0)

        # train tree models for every objective
        for i, obj in enumerate(self._problem_config.obj_list):
            if self._train_lib == "lgbm":
                tree_model = self._train_lgbm(X, y[:, i])
            elif self._train_lib == "catboost":
                raise NotImplementedError()
            elif self._train_lib == "xgboost":
                raise NotImplementedError()
            else:
                raise IOError(
                    "Parameter 'train_lib' for tree ensembles needs to be "
                    "in '('lgbm', 'catboost', 'xgboost')'."
                )
            self._tree_dict[obj.name] = tree_model

    def _train_lgbm(self, X, y):
        import lightgbm as lgb

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            if self._problem_config.cat_idx:
                # train for categorial vars
                train_data = lgb.Dataset(
                    X,
                    label=y,
                    categorical_feature=list(self._problem_config.cat_idx),
                    free_raw_data=False,
                    params={"verbose": -1},
                )

                tree_model = lgb.train(
                    self._train_params,
                    train_data,
                    #verbose_eval=False,
                )
            else:
                # train for non-categorical vars
                train_data = lgb.Dataset(X, label=y, params={"verbose": -1})

                tree_model = lgb.train(
                    self._train_params, train_data#, verbose_eval=False
                )
        return tree_model

    def predict(self, X):
        # predict vals
        tree_pred = []
        for obj in self._problem_config.obj_list:
            tree_pred.append(self.tree_dict[obj.name].predict(X))
        return np.column_stack(tree_pred)

    def _update_meta_tree_dict(self):
        self._meta_tree_dict = {}

        # get model information
        for obj in self._problem_config.obj_list:
            if self._train_lib == "lgbm":
                lib_out = self.tree_dict[obj.name].dump_model()
            elif self._train_lib == "catboost":
                raise NotImplementedError()
            elif self._train_lib == "xgboost":
                raise NotImplementedError()
            else:
                raise IOError(
                    "Parameter 'train_lib' for tree ensembles needs to be "
                    "in '('lgbm')'."
                )

            # order tree_model_dict
            ordered_tree_model_dict = read_lgbm_tree_model_dict(
                lib_out, cat_idx=self._problem_config.cat_idx
            )

            # populate meta_tree_model
            self._meta_tree_dict[obj.name] = MetaTreeModel(ordered_tree_model_dict)

    def add_to_gurobipy_model(self, model, add_mu_var=True, normalize_mean=False):
        from gurobipy import GRB, quicksum

        self._update_meta_tree_dict()

        # attach tree info
        model._num_trees = lambda obj_name: self._meta_tree_dict[obj_name].num_trees

        # attach leaf info
        model._leaves = lambda obj_name, tree: tuple(
            self._meta_tree_dict[obj_name].get_leaf_encodings(tree)
        )
        model._leaf_weight = lambda obj_name, tree, leaf: self._meta_tree_dict[
            obj_name
        ].get_leaf_weight(tree, leaf)
        model._leaf_vars = lambda obj_name, tree, leaf: tuple(
            var
            for var in self._meta_tree_dict[obj_name].get_participating_variables(
                tree, leaf
            )
        )

        # attach breakpoint info
        var_break_points = [
            tree_model.get_var_break_points()
            for tree_model in self._meta_tree_dict.values()
        ]
        model._breakpoints = {}
        model._breakpoint_index = []

        for idx, feat in enumerate(self._problem_config.feat_list):
            if feat.is_cat():
                continue
            else:
                splits = set()
                for vb in var_break_points:
                    try:
                        splits = splits.union(set(vb[idx]))
                    except KeyError:
                        pass
                if splits:
                    model._breakpoints[idx] = sorted(splits)
                    model._breakpoint_index.append(idx)

        # define indexing helper functions
        def tree_index(model_obj):
            for obj in self._problem_config.obj_list:
                for tree in range(model_obj._num_trees(obj.name)):
                    yield obj.name, tree

        def leaf_index(model_obj):
            for label, tree in tree_index(model_obj):
                for leaf in model_obj._leaves(label, tree):
                    yield label, tree, leaf

        def interval_index(model_obj):
            for var in model_obj._breakpoint_index:
                for j in range(len(model_obj._breakpoints[var])):
                    yield var, j

        def split_index(model_obj, meta_tree_dict):
            for label, tree in tree_index(model_obj):
                for encoding in meta_tree_dict[label].get_branch_encodings(tree):
                    yield label, tree, encoding

        # add leaf variables
        model._z = model.addVars(
            leaf_index(model), lb=0, ub=GRB.INFINITY, name="z", vtype="C"
        )

        # add split variables
        model._nu = model.addVars(interval_index(model), name="nu", vtype=GRB.BINARY)

        # add single leaf constraints
        def single_leaf_rule(model_obj, label, tree):
            z, leaves = model_obj._z, model_obj._leaves
            return quicksum(z[label, tree, leaf] for leaf in leaves(label, tree)) == 1

        model.addConstrs(
            (
                single_leaf_rule(model, label, tree)
                for (label, tree) in tree_index(model)
            ),
            name="single_leaf",
        )

        # add left split constraints
        def left_split_r(model_obj, meta_tree_dict, label, tree, split_enc):
            split_var, split_val = meta_tree_dict[label].get_branch_partition_pair(
                tree, split_enc
            )

            if not isinstance(split_val, list):
                # non-cat vars
                nu_val = model_obj._breakpoints[split_var].index(split_val)
                return (
                    quicksum(
                        model_obj._z[label, tree, leaf]
                        for leaf in meta_tree_dict[label].get_left_leaves(
                            tree, split_enc
                        )
                    )
                    <= model_obj._nu[split_var, nu_val]
                )
            else:
                # cat vars
                return quicksum(
                    model_obj._z[label, tree, leaf]
                    for leaf in meta_tree_dict[label].get_left_leaves(tree, split_enc)
                ) <= quicksum(model_obj._all_feat[split_var][cat] for cat in split_val)

        model.addConstrs(
            (
                left_split_r(model, self.meta_tree_dict, label, tree, encoding)
                for (label, tree, encoding) in split_index(model, self.meta_tree_dict)
            ),
            name="left_split",
        )

        # add right split constraints
        def right_split_r(model_obj, meta_tree_dict, label, tree, split_enc):
            split_var, split_val = meta_tree_dict[label].get_branch_partition_pair(
                tree, split_enc
            )

            if not isinstance(split_val, list):
                # non-cat vars
                nu_val = model_obj._breakpoints[split_var].index(split_val)
                return (
                    quicksum(
                        model_obj._z[label, tree, leaf]
                        for leaf in meta_tree_dict[label].get_right_leaves(
                            tree, split_enc
                        )
                    )
                    <= 1 - model_obj._nu[split_var, nu_val]
                )
            else:
                # cat vars
                return quicksum(
                    model_obj._z[label, tree, leaf]
                    for leaf in meta_tree_dict[label].get_right_leaves(tree, split_enc)
                ) <= 1 - quicksum(
                    model_obj._all_feat[split_var][cat] for cat in split_val
                )

        model.addConstrs(
            (
                right_split_r(model, self.meta_tree_dict, label, tree, encoding)
                for (label, tree, encoding) in split_index(model, self.meta_tree_dict)
            ),
            name="right_split",
        )

        # add split order constraints
        def y_order_r(model_obj, i, j):
            return model_obj._nu[i, j] <= model_obj._nu[i, j + 1]

        model.addConstrs(
            (
                y_order_r(model, var, j)
                for (var, j) in interval_index(model)
                if j != len(model._breakpoints[var]) - 1
            ),
            name="y_order",
        )

        # add cat var constraints
        def cat_sums(model_obj, i):
            return (
                quicksum(model_obj._all_feat[i][cat] for cat in model_obj._all_feat[i])
                == 1
            )

        model.addConstrs(
            (cat_sums(model, var) for var in self._problem_config.cat_idx),
            name="cat_sums",
        )

        # add split / space linking constraints
        def var_lower_r(model_obj, lb, i, j):
            j_bound = model_obj._breakpoints[i][j]
            return model_obj._all_feat[i] >= lb + (j_bound - lb) * (
                1 - model_obj._nu[i, j]
            )

        def var_upper_r(model_obj, ub, i, j):
            j_bound = model_obj._breakpoints[i][j]
            return model_obj._all_feat[i] <= ub + (j_bound - ub) * (model_obj._nu[i, j])

        model.addConstrs(
            (
                var_lower_r(model, self._problem_config.feat_list[var].lb, var, j)
                for (var, j) in interval_index(model)
            ),
            name="var_lower",
        )

        model.addConstrs(
            (
                var_upper_r(model, self._problem_config.feat_list[var].ub, var, j)
                for (var, j) in interval_index(model)
            ),
            name="var_upper",
        )

        if add_mu_var:
            # add mu variables for all objectives
            def obj_leaf_index(model_obj, obj_name):
                for tree in range(model_obj._num_trees(obj_name)):
                    for leaf in model_obj._leaves(obj_name, tree):
                        yield tree, leaf

            model._aux_mu = []
            model._unscaled_mu = []

            for idx, obj in enumerate(self._problem_config.obj_list):
                weighted_sum = quicksum(
                    model._leaf_weight(obj.name, tree, leaf)
                    * model._z[obj.name, tree, leaf]
                    for tree, leaf in obj_leaf_index(model, obj.name)
                )

                model._aux_mu.append(
                    model.addVar(
                        lb=-GRB.INFINITY,
                        ub=GRB.INFINITY,
                        name=f"aux_mean_obj_{obj.name}",
                        vtype="C",
                    )
                )

                model._unscaled_mu.append(
                    model.addVar(
                        lb=-GRB.INFINITY,
                        ub=GRB.INFINITY,
                        name=f"unscaled_mean_obj_{obj.name}",
                        vtype="C",
                    )
                )

                if normalize_mean:
                    shift, scale = self.min_y[idx], self.max_y[idx] - self.min_y[idx]
                else:
                    shift, scale = 0.0, 1.0

                model.addConstr(
                    model._aux_mu[-1] == (weighted_sum - shift) / scale,
                    name=f"aux_mean_obj_{obj.name}_tree_link",
                )

                model.addConstr(
                    model._unscaled_mu[-1] == (model._aux_mu[-1] * scale) + shift,
                    name=f"unscaled_mean_obj_{obj.name}_tree_link",
                )

        model.update()

    def add_to_pyomo_model(
        self, model, add_mu_var: bool = True, normalize_mean: bool = False
    ):
        import pyomo.environ as pyo

        self._update_meta_tree_dict()

        # attach tree info
        model._num_trees = lambda obj_name: self._meta_tree_dict[obj_name].num_trees

        # attach leaf info
        model._leaves = lambda obj_name, tree: tuple(
            self._meta_tree_dict[obj_name].get_leaf_encodings(tree)
        )
        model._leaf_weight = lambda obj_name, tree, leaf: self._meta_tree_dict[
            obj_name
        ].get_leaf_weight(tree, leaf)
        model._leaf_vars = lambda obj_name, tree, leaf: tuple(
            var
            for var in self._meta_tree_dict[obj_name].get_participating_variables(
                tree, leaf
            )
        )

        # attach breakpoint info
        var_break_points = [
            tree_model.get_var_break_points()
            for tree_model in self._meta_tree_dict.values()
        ]
        model._breakpoints = {}
        model._breakpoint_index = []

        for idx, feat in enumerate(self._problem_config.feat_list):
            if feat.is_cat():
                continue
            else:
                splits = set()
                for vb in var_break_points:
                    try:
                        splits = splits.union(set(vb[idx]))
                    except KeyError:
                        pass
                if splits:
                    model._breakpoints[idx] = sorted(splits)
                    model._breakpoint_index.append(idx)

        # define indexing helper functions
        def tree_index(model_obj):
            for obj in self._problem_config.obj_list:
                for tree in range(model_obj._num_trees(obj.name)):
                    yield obj.name, tree

        def leaf_index(model_obj):
            for label, tree in tree_index(model_obj):
                for leaf in model_obj._leaves(label, tree):
                    yield label, tree, leaf

        def interval_index(model_obj):
            for var in model_obj._breakpoint_index:
                for j in range(len(model_obj._breakpoints[var])):
                    yield var, j

        def split_index(model_obj, meta_tree_dict):
            for label, tree in tree_index(model_obj):
                for encoding in meta_tree_dict[label].get_branch_encodings(tree):
                    yield label, tree, encoding

        # add leaf variables
        model._z = pyo.Var(leaf_index(model), domain=pyo.NonNegativeReals)

        # add split variables
        model._nu = pyo.Var(interval_index(model), domain=pyo.Binary)

        def single_leaf_rule(model_obj, label, tree):
            z, leaves = model_obj._z, model_obj._leaves
            return sum(z[label, tree, leaf] for leaf in leaves(label, tree)) == 1

        model.single_leaf_constraints = pyo.Constraint(
            [(label, tree) for (label, tree) in tree_index(model)],
            rule=single_leaf_rule,
        )

        # We hide the dictionary in the model data since we will need it some functions (e.g.left_split_r()), but
        # the function is not allowed to have the dict as an argument since otherwise Pyomo would not accept it as rule.
        model.meta_tree_dict = self.meta_tree_dict

        # add left split constraints
        def left_split_r(model_obj, label, tree, split_enc):
            meta_tree_dict = model.meta_tree_dict
            split_var, split_val = meta_tree_dict[label].get_branch_partition_pair(
                tree, split_enc
            )

            if not isinstance(split_val, list):
                # non-cat vars
                nu_val = model_obj._breakpoints[split_var].index(split_val)
                return (
                    sum(
                        model_obj._z[label, tree, leaf]
                        for leaf in meta_tree_dict[label].get_left_leaves(
                            tree, split_enc
                        )
                    )
                    <= model_obj._nu[split_var, nu_val]
                )
            else:
                # cat vars
                return sum(
                    model_obj._z[label, tree, leaf]
                    for leaf in meta_tree_dict[label].get_left_leaves(tree, split_enc)
                ) <= sum(model_obj._all_feat[split_var][cat] for cat in split_val)

        model.left_split_constraints = pyo.Constraint(
            [
                (label, tree, encoding)
                for (label, tree, encoding) in split_index(model, self.meta_tree_dict)
            ],
            rule=left_split_r,
        )

        # add right split constraints
        def right_split_r(model_obj, label, tree, split_enc):
            meta_tree_dict = model_obj.meta_tree_dict
            split_var, split_val = meta_tree_dict[label].get_branch_partition_pair(
                tree, split_enc
            )

            if not isinstance(split_val, list):
                # non-cat vars
                nu_val = model_obj._breakpoints[split_var].index(split_val)
                return (
                    sum(
                        model_obj._z[label, tree, leaf]
                        for leaf in meta_tree_dict[label].get_right_leaves(
                            tree, split_enc
                        )
                    )
                    <= 1 - model_obj._nu[split_var, nu_val]
                )
            else:
                # cat vars
                return sum(
                    model_obj._z[label, tree, leaf]
                    for leaf in meta_tree_dict[label].get_right_leaves(tree, split_enc)
                ) <= 1 - sum(model_obj._all_feat[split_var][cat] for cat in split_val)

        model.right_split_constraints = pyo.Constraint(
            [
                (label, tree, encoding)
                for (label, tree, encoding) in split_index(model, self.meta_tree_dict)
            ],
            rule=right_split_r,
        )

        # add split order constraints
        def y_order_r(model_obj, i, j):
            return model_obj._nu[i, j] <= model_obj._nu[i, j + 1]

        model.y_order_constraints = pyo.Constraint(
            [
                (var, j)
                for (var, j) in interval_index(model)
                if j != len(model._breakpoints[var]) - 1
            ],
            rule=y_order_r,
        )

        # add cat var constraints
        def cat_sums(model_obj, i):
            return (
                sum(model_obj._all_feat[i][cat] for cat in model_obj._all_feat[i]) == 1
            )

        model.categorical_sum_constraints = pyo.Constraint(
            [var for var in self._problem_config.cat_idx], rule=cat_sums
        )

        # add split / space linking constraints
        def var_lower_r(model_obj, i, j):
            lb = model_obj._all_feat[i].lb
            j_bound = model_obj._breakpoints[i][j]
            return model_obj._all_feat[i] >= lb + (j_bound - lb) * (
                1 - model_obj._nu[i, j]
            )

        def var_upper_r(model_obj, i, j):
            ub = model_obj._all_feat[i].ub
            j_bound = model_obj._breakpoints[i][j]
            return model_obj._all_feat[i] <= ub + (j_bound - ub) * (model_obj._nu[i, j])

        model.linking_constraints_lower = pyo.Constraint(
            [(var, j) for (var, j) in interval_index(model)], rule=var_lower_r
        )

        model.linking_constraints_upper = pyo.Constraint(
            [(var, j) for (var, j) in interval_index(model)], rule=var_upper_r
        )

        if add_mu_var:
            # add mu variables for all objectives
            def obj_leaf_index(model_obj, obj_name):
                for tree in range(model_obj._num_trees(obj_name)):
                    for leaf in model_obj._leaves(obj_name, tree):
                        yield tree, leaf

            objective_names = [obj.name for obj in self._problem_config.obj_list]

            model._aux_mu = pyo.Var(objective_names, domain=pyo.Reals)
            model._unscaled_mu = pyo.Var(objective_names, domain=pyo.Reals)

            weighted_sum = {}
            shift = {}
            scale = {}

            for idx, obj_name in enumerate(objective_names):
                weighted_sum[obj_name] = sum(
                    model._leaf_weight(obj_name, tree, leaf)
                    * model._z[obj_name, tree, leaf]
                    for tree, leaf in obj_leaf_index(model, obj_name)
                )

                if normalize_mean:
                    shift[obj_name], scale[obj_name] = (
                        self.min_y[idx],
                        self.max_y[idx] - self.min_y[idx],
                    )
                else:
                    shift[obj_name], scale[obj_name] = 0.0, 1.0

            def mu_objectives(model_obj, name):
                return (
                    model_obj._aux_mu[name]
                    == (weighted_sum[name] - shift[name]) / scale[name]
                )

            model.constraints_mu_objectives = pyo.Constraint(
                objective_names, rule=mu_objectives
            )

            def link_unscaled_mu_auxmu(model_obj: pyo.ConcreteModel, name: str):
                return (
                    model_obj._unscaled_mu[name]
                    == (model._aux_mu[name] * scale[name]) + shift[name]
                )

            model.constraints_link_unscaled_mu_auxmu = pyo.Constraint(
                objective_names, rule=link_unscaled_mu_auxmu
            )



from entmoot.models.base_model import BaseModel
from entmoot.models.uncertainty_models.base_distance import CatDistance, NonCatDistance
from entmoot.models.uncertainty_models.euclidean_squared_distance import (
    EuclideanSquaredDistance,
)
from entmoot.models.uncertainty_models.l1_distance import L1Distance
from entmoot.models.uncertainty_models.l2_distance import L2Distance

from entmoot.models.uncertainty_models.overlap_distance import OverlapDistance
from entmoot.models.uncertainty_models.goodall4_distance import Goodall4Distance
from entmoot.models.uncertainty_models.of_distance import OfDistance

from entmoot.models.model_params import UncParams, ParamValidationError
from typing import Union
import numpy as np

def distance_func_mapper(dist_name: str, cat: bool) -> Union[CatDistance, NonCatDistance]:
    """Given a string, return the distance function"""
    non_cat_dists = {
        "euclidean_squared": EuclideanSquaredDistance,
        "l1": L1Distance,
        "l2": L2Distance,
    }
    cat_dists = {
        "overlap": OverlapDistance,
        "of": OfDistance,
        "goodall4": Goodall4Distance,
    }
    if cat:
        return cat_dists.get(dist_name)
    else:
        return non_cat_dists.get(dist_name)


class DistanceBasedUncertainty(BaseModel):
    def __init__(self, problem_config, params: Union[UncParams, dict, None] = None):
        if params is None:
            params = {}
        if isinstance(params, dict):
            params = UncParams(**params)

        self._problem_config = problem_config

        self._non_cat_x, self._cat_x = None, None
        self._dist_bound = None
        self._dist_metric = params.dist_metric
        self._num_cache_x = None
        self._acq_sense = params.acq_sense

        if params.dist_trafo == "standard":
            assert (
                len(self._problem_config.obj_list) == 1
            ), "Distance transformation 'standard' can only be used for single objective problems."

            assert (
                len(self._problem_config.cat_idx) == 0
            ), "Distance transformation 'standard' can only be used for non-categorical problems."

            self._dist_has_var_bound = False if self._acq_sense == "penalty" else True
            self._bound_coeff = params.bound_coeff
            self._dist_coeff = 1.0
        elif params.dist_trafo == "normal":
            self._dist_has_var_bound = False
            self._bound_coeff = None
            self._dist_coeff = 1 / len(self._problem_config.feat_list)
        else:
            raise ParamValidationError(
                f"Pick 'dist_trafo' '{params.dist_trafo}' in '('normal', 'standard')'."
            )

        # pick distance metric for non-cat features
        non_cat_distance = distance_func_mapper(params.dist_metric, cat=False)
        if non_cat_distance is None:
            raise ParamValidationError(
                f"Non-categorical uncertainty metric '{params.dist_metric}' for "
                f"{self.__class__.__name__} model is not supported. "
                f"Check 'params['dist_metric']'."
            )
        else:
            self.non_cat_unc_model: NonCatDistance = non_cat_distance(
                problem_config=self._problem_config,
                acq_sense=params.acq_sense,
                dist_trafo=params.dist_trafo,
            )

        # pick distance metric for cat features
        cat_distance = distance_func_mapper(params.cat_metric, cat=True)
        if cat_distance is None:
            raise ParamValidationError(
                f"Categorical uncertainty metric '{params.cat_metric}' for {self.__class__.__name__} "
                f"model is not supported. Check 'params['cat_metric']'."
            )
        else:
            self.cat_unc_model: CatDistance = cat_distance(
                problem_config=self._problem_config, 
                acq_sense=params.acq_sense,
            )

    @property
    def num_cache_x(self):
        assert (
            self._num_cache_x is not None
        ), f"Uncertainty model needs fit function call before it can predict."
        return self._num_cache_x

    def fit(self, X, y):
        if self._dist_has_var_bound:
            self._dist_bound = abs(np.var(y) * self._bound_coeff)

        self._num_cache_x = len(X)

        self.non_cat_unc_model.fit(X)
        self.cat_unc_model.fit(X)

    def predict(self, X):
        comb_pred = []
        for xi in X:
            non_cat_pred = self.non_cat_unc_model.predict(xi)
            cat_pred = self.cat_unc_model.predict(xi)
            dist_pred = np.min(non_cat_pred + cat_pred) * self._dist_coeff

            # the standard trafo case has a bound on the prediction
            if self._dist_has_var_bound:
                if dist_pred > self._dist_bound:
                    dist_pred = self._dist_bound

            comb_pred.append(dist_pred)
        return np.asarray(comb_pred)

    def add_to_gurobipy_model(self, model):
        from gurobipy import GRB, quicksum

        # define main uncertainty variables
        if self._dist_has_var_bound:
            dist_bound = self._dist_bound
        else:
            dist_bound = GRB.INFINITY

        model._unc = model.addVar(lb=0.0, ub=dist_bound, name="uncertainty", vtype="C")

        # get constr terms for non-categorical and categorical contributions
        non_cat_term_list = self.non_cat_unc_model.get_gurobipy_model_constr_terms(
            model
        )
        cat_term_list = self.cat_unc_model.get_gurobipy_model_constr_terms(model)

        # add binaries for big_m penalty constraints
        if self._acq_sense == "penalty":
            model._bin_penalty = []
            big_m = self.non_cat_unc_model.get_big_m() + self.cat_unc_model.get_big_m()

        for i, (non_cat_term, cat_term) in enumerate(
            zip(non_cat_term_list, cat_term_list)
        ):
            # check if penalty term is needed
            if self._acq_sense == "penalty":
                model._bin_penalty.append(
                    model.addVar(name=f"bin_penalty_{i}", vtype="B")
                )

                big_m_term = big_m * (1 - model._bin_penalty[-1])

                if self._dist_metric == "l2":
                    # take sqrt for l2 distance
                    aux_non_cat_unc = model.addVar(
                        lb=0.0, ub=dist_bound, name=f"aux_non_cat_unc_x_{i}", vtype="C"
                    )

                    model.addQConstr(
                        aux_non_cat_unc * aux_non_cat_unc >= non_cat_term,
                        name=f"unc_x_{i}",
                    )

                    model.addQConstr(
                        model._unc + big_m_term
                        >= (aux_non_cat_unc + cat_term) * self._dist_coeff,
                        name=f"unc_x_{i}",
                    )
                    model.params.NonConvex = 2
                else:
                    # self._dist_metric == "l1" or "euclidean_squared"
                    model.addQConstr(
                        model._unc + big_m_term
                        >= (non_cat_term + cat_term) * self._dist_coeff,
                        name=f"unc_x_{i}",
                    )

            elif self._acq_sense == "exploration":
                if self._dist_metric == "l2":
                    # take sqrt for l2 distance
                    aux_non_cat_unc = model.addVar(
                        lb=0.0, ub=dist_bound, name=f"aux_non_cat_unc_x_{i}", vtype="C"
                    )

                    model.addQConstr(
                        aux_non_cat_unc * aux_non_cat_unc <= non_cat_term,
                        name=f"unc_x_{i}",
                    )

                    model.addQConstr(
                        model._unc <= (aux_non_cat_unc + cat_term) * self._dist_coeff,
                        name=f"unc_x_{i}",
                    )
                else:
                    model.addQConstr(
                        model._unc <= (non_cat_term + cat_term) * self._dist_coeff,
                        name=f"unc_x_{i}",
                    )

        if self._acq_sense == "exploration":
            model.params.NonConvex = 2

        if self._acq_sense == "penalty":
            model.addConstr(sum(model._bin_penalty) == 1, name=f"bin_penalty_sum")

        model.update()

    def add_to_pyomo_model(self, model):
        import pyomo.environ as pyo

        # define main uncertainty variables
        if self._dist_has_var_bound:
            dist_bound = self._dist_bound
        else:
            dist_bound = float("inf")

        model._unc = pyo.Var(bounds=(0, dist_bound), domain=pyo.Reals)

        # get constr terms for non-categorical and categorical contributions
        non_cat_term_list = self.non_cat_unc_model.get_pyomo_model_constr_terms(model)
        cat_term_list = self.cat_unc_model.get_pyomo_model_constr_terms(model)

        model.terms_constrs_cat_noncat_contr = list(
            zip(non_cat_term_list, cat_term_list)
        )
        model.indices_constrs_cat_noncat_contr = list(
            range(len(model.terms_constrs_cat_noncat_contr))
        )

        if self._acq_sense == "penalty":
            big_m = self.non_cat_unc_model.get_big_m() + self.cat_unc_model.get_big_m()

            # add binaries for big_m penalty constraints
            model._bin_penalty = pyo.Var(
                model.indices_constrs_cat_noncat_contr, domain=pyo.Binary
            )

            big_m_term = {
                i: big_m * (1 - model._bin_penalty[i])
                for i in model.indices_constrs_cat_noncat_contr
            }

            def constrs_bin_penalty_sum(model_obj):
                return (
                    sum(model_obj._bin_penalty[k] for k in model.indices_constrs_cat_noncat_contr) == 1
                )

            model.constrs_bin_penalty_sum = pyo.Constraint(rule=constrs_bin_penalty_sum)

            if self._dist_metric == "l2":
                # take sqrt for l2 distance
                model.aux_non_cat_unc = pyo.Var(
                    model.indices_constrs_cat_noncat_contr, bounds=(0, dist_bound)
                )

                def constrs_unc_x_aux(model_obj, k):
                    return (
                        model_obj.aux_non_cat_unc[k] * model_obj.aux_non_cat_unc[k]
                        >= non_cat_term_list[k]
                    )

                model.constrs_unc_x_aux = pyo.Constraint(
                    model.indices_constrs_cat_noncat_contr, rule=constrs_unc_x_aux
                )

                def constrs_unc_x(model_obj, k):
                    return (
                        model_obj._unc + big_m_term[k]
                        >= (model_obj.aux_non_cat_unc[k] + cat_term_list[k])
                        * self._dist_coeff
                    )

                model.constrs_unc_x_l2 = pyo.Constraint(
                    model.indices_constrs_cat_noncat_contr, rule=constrs_unc_x
                )

            elif self._dist_metric == "l1" or "euclidean_squared":

                def constrs_unc_x_l1l2squared(model_obj, k):
                    return (
                        model_obj._unc + big_m_term[k]
                        >= (non_cat_term_list[k] + cat_term_list[k]) * self._dist_coeff
                    )

                model.constrs_unc_x_l1l2squared = pyo.Constraint(
                    model.indices_constrs_cat_noncat_contr,
                    rule=constrs_unc_x_l1l2squared,
                )

            else:
                # Illegal value for self._dist_metric
                raise ValueError
        elif self._acq_sense == "exploration":
            if self._dist_metric == "l2":
                # take sqrt for l2 distance
                model.aux_non_cat_unc = pyo.Var(
                    model.indices_constrs_cat_noncat_contr, bounds=(0, dist_bound)
                )

                def constrs_unc_x_aux_expl(model_obj, k):
                    return (
                        model_obj.aux_non_cat_unc[k] * model_obj.aux_non_cat_unc[k]
                        <= non_cat_term_list[k]
                    )

                model.constrs_unc_x_aux_expl = pyo.Constraint(
                    model.indices_constrs_cat_noncat_contr, rule=constrs_unc_x_aux_expl
                )

                def constrs_unc_x_expl(model_obj, k):
                    return (
                        model_obj._unc
                        <= (model_obj.aux_non_cat_unc[k] + cat_term_list[k])
                        * self._dist_coeff
                    )

                model.constrs_unc_x_expl = pyo.Constraint(
                    model.indices_constrs_cat_noncat_contr, rule=constrs_unc_x_expl
                )

            elif self._dist_metric == "l1" or "euclidean_squared":

                def constrs_unc_x_l1l2squared_expl(model_obj, k):
                    return (
                        model_obj._unc
                        <= (non_cat_term_list[k] + cat_term_list[k]) * self._dist_coeff
                    )

                model.constrs_unc_x_l1l2squared = pyo.Constraint(
                    model.indices_constrs_cat_noncat_contr,
                    rule=constrs_unc_x_l1l2squared_expl,
                )

            else:
                # Illegal value for self._dist_metric
                raise ValueError

        def rule_constr_cat_noncat_quadr(model_obj, i):
            # take sqrt for l2 distance
            non_cat_term, cat_term = model.terms_constrs_cat_noncat_contr[i]
            return (
                model._unc * model._unc
                <= (non_cat_term + cat_term) * self._dist_coeff**2
            )

        def rule_constr_cat_noncat(model_obj, i):
            non_cat_term, cat_term = model.terms_constrs_cat_noncat_contr[i]
            return model._unc <= (non_cat_term + cat_term) * self._dist_coeff

        if not self._acq_sense == "exploration":
            if self._dist_metric == "l2":
                model.constrs_cat_noncat_contr = pyo.Constraint(
                    model.indices_constrs_cat_noncat_contr,
                    rule=rule_constr_cat_noncat_quadr,
                )
            else:
                model.constrs_cat_noncat_contr = pyo.Constraint(
                    model.indices_constrs_cat_noncat_contr, rule=rule_constr_cat_noncat
                )
